import sys
import warnings
from abc import ABC
from pathlib import Path
from typing import ClassVar

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator
from rdflib import DCTERMS, OWL, RDF, RDFS, XSD, BNode, Graph, Literal, Namespace, URIRef
from rdflib.collection import Collection as GraphCollection

from cognite.neat.constants import DEFAULT_NAMESPACE as NEAT_NAMESPACE
from cognite.neat.rules import exceptions
from cognite.neat.rules._analysis._information_rules import InformationArchitectRulesAnalysis
from cognite.neat.rules.models._rules import DMSRules
from cognite.neat.rules.models._rules._types import XSD_VALUE_TYPE_MAPPINGS, EntityTypes
from cognite.neat.rules.models._rules.information_rules import (
    InformationClass,
    InformationMetadata,
    InformationProperty,
    InformationRules,
)
from cognite.neat.utils.utils import generate_exception_report, remove_namespace

from ._base import BaseExporter
from ._validation import are_properties_redefined

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from cognite.neat.rules._shared import Rules


class GraphExporter(BaseExporter[Graph], ABC):
    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        self.export(rules).serialize(destination=filepath, encoding=self._encoding, newline=self._new_line)


class OWLExporter(GraphExporter):
    def export(self, rules: Rules) -> Graph:
        return Ontology.from_rules(rules).as_owl()


class SHACLExporter(GraphExporter):
    def export(self, rules: Rules) -> Graph:
        return Ontology.from_rules(rules).as_shacl()


class SemanticDataModelExporter(GraphExporter):
    def export(self, rules: Rules) -> Graph:
        return Ontology.from_rules(rules).as_semantic_data_model()


class OntologyModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True, strict=False, extra="allow")


class Ontology(OntologyModel):
    """
    Represents an ontology. Thi class is used to generate an OWL ontology from a set of transformation rules.

    Args:
        properties: A list of OWL properties.
        classes: A list of OWL classes.
        shapes: A list of SHACL node shapes.
        metadata: Metadata about the ontology.
        prefixes: A dictionary of prefixes and namespaces.
    """

    properties: list["OWLProperty"]
    classes: list["OWLClass"]
    shapes: list["SHACLNodeShape"]
    metadata: "OWLMetadata"
    prefixes: dict[str, Namespace]

    @classmethod
    def from_rules(cls, input_rules: Rules) -> Self:
        """
        Generates an ontology from a set of transformation rules.

        Args:
            input_rules: The rules to generate the ontology from.

        Returns:
            An instance of Ontology.
        """
        if isinstance(input_rules, InformationRules):
            rules = input_rules
        elif isinstance(input_rules, DMSRules):
            rules = input_rules.as_information_architect_rules()
        else:
            raise ValueError(f"{type(input_rules).__name__} cannot be exported to Ontology")

        properties_redefined, redefinition_warnings = are_properties_redefined(rules, return_report=True)
        if properties_redefined:
            raise exceptions.PropertiesDefinedMultipleTimes(report=generate_exception_report(redefinition_warnings))

        if rules.prefixes is None:
            raise exceptions.PrefixMissing()

        if rules.metadata.namespace is None:
            raise exceptions.MissingDataModelPrefixOrNamespace()

        class_dict = InformationArchitectRulesAnalysis(rules).as_class_dict()
        return cls(
            properties=[
                OWLProperty.from_list_of_properties(definition, rules.metadata.namespace)
                for definition in InformationArchitectRulesAnalysis(rules).as_property_dict().values()
            ],
            classes=[
                OWLClass.from_class(definition, rules.metadata.namespace, rules.prefixes)
                for definition in rules.classes
            ],
            shapes=[
                SHACLNodeShape.from_rules(
                    class_dict[class_.suffix],
                    list(properties.values()),
                    rules.metadata.namespace,
                )
                for class_, properties in InformationArchitectRulesAnalysis(rules).class_property_pairs().items()
            ]
            + [
                SHACLNodeShape.from_rules(
                    class_,
                    [],
                    rules.metadata.namespace,
                )
                for class_ in class_dict.values()
            ],
            metadata=OWLMetadata(**rules.metadata.model_dump()),
            prefixes=rules.prefixes,
        )

    def as_shacl(self) -> Graph:
        """
        Generates a SHACL graph from the ontology.

        Returns:
            A SHACL graph.
        """

        shacl = Graph()
        shacl.bind(self.metadata.prefix, self.metadata.namespace)
        for prefix, namespace in self.prefixes.items():
            shacl.bind(prefix, namespace)

        for shape in self.shapes:
            for triple in shape.triples:
                shacl.add(triple)  # type: ignore[arg-type]

        return shacl

    def as_owl(self) -> Graph:
        """
        Generates an OWL graph from the ontology.

        Returns:
            An OWL graph.
        """
        owl = Graph()
        owl.bind(self.metadata.prefix, self.metadata.namespace)
        for prefix, namespace in self.prefixes.items():
            owl.bind(prefix, namespace)

        if self.metadata.namespace is None:
            raise exceptions.MetadataSheetNamespaceNotDefined()

        owl.add((URIRef(self.metadata.namespace), RDF.type, OWL.Ontology))
        for property_ in self.properties:
            for triple in property_.triples:
                owl.add(triple)  # type: ignore[arg-type]

        for class_ in self.classes:
            for triple in class_.triples:
                owl.add(triple)  # type: ignore[arg-type]

        for triple in self.metadata.triples:
            owl.add(triple)  # type: ignore[arg-type]

        return owl

    def as_semantic_data_model(self) -> Graph:
        return self.as_owl() + self.as_shacl()

    @property
    def owl_triples(self) -> list[tuple]:
        return list(self.as_owl())

    @property
    def shacl_triples(self) -> list[tuple]:
        return list(self.as_shacl())

    @property
    def triples(self) -> list[tuple]:
        return self.owl_triples + self.shacl_triples

    @property
    def ontology(self) -> str:
        return self.as_owl().serialize()

    @property
    def constraints(self) -> str:
        return self.as_shacl().serialize()

    @property
    def semantic_data_model(self) -> str:
        return (self.as_owl() + self.as_shacl()).serialize()


class OWLMetadata(InformationMetadata):
    @property
    def triples(self) -> list[tuple]:
        # Mandatory triples originating from Metadata mandatory fields
        if self.namespace is None:
            raise exceptions.MetadataSheetNamespaceNotDefined()
        triples: list[tuple] = [
            (URIRef(self.namespace), DCTERMS.hasVersion, Literal(self.version)),
            (URIRef(self.namespace), OWL.versionInfo, Literal(self.version)),
            (URIRef(self.namespace), RDFS.label, Literal(self.name)),
            (URIRef(self.namespace), NEAT_NAMESPACE.prefix, Literal(self.prefix)),
            (URIRef(self.namespace), DCTERMS.title, Literal(self.name)),
            (URIRef(self.namespace), DCTERMS.created, Literal(self.created, datatype=XSD.dateTime)),
            (URIRef(self.namespace), DCTERMS.description, Literal(self.description)),
        ]
        if isinstance(self.creator, list):
            triples.extend([(URIRef(self.namespace), DCTERMS.creator, Literal(creator)) for creator in self.creator])
        else:
            triples.append((URIRef(self.namespace), DCTERMS.creator, Literal(self.creator)))

        # Optional triples originating from Metadata optional fields
        if self.updated:
            triples.append((URIRef(self.namespace), DCTERMS.modified, Literal(self.updated, datatype=XSD.dateTime)))

        return triples


class OWLClass(OntologyModel):
    id_: URIRef
    type_: URIRef = OWL.Class
    label: str | None
    comment: str | None
    sub_class_of: list[URIRef] | None
    namespace: Namespace

    @classmethod
    def from_class(cls, definition: InformationClass, namespace: Namespace, prefixes: dict) -> Self:
        if definition.parent and isinstance(definition.parent, list):
            sub_class_of = []
            for parent_class in definition.parent:
                try:
                    sub_class_of.append(prefixes[parent_class.prefix][parent_class.suffix])
                except KeyError:
                    sub_class_of.append(namespace[parent_class.suffix])
        else:
            sub_class_of = None

        return cls(
            id_=namespace[definition.class_.suffix],
            label=definition.name or definition.class_.suffix,
            comment=definition.description,
            sub_class_of=sub_class_of,
            namespace=namespace,
        )

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, self.type_)]

    @property
    def label_triples(self) -> list[tuple]:
        if self.label:
            return [(self.id_, RDFS.label, Literal(self.label))]
        else:  # If comment is None, return empty list
            return []

    @property
    def comment_triples(self) -> list[tuple]:
        if self.comment:
            return [(self.id_, RDFS.comment, Literal(self.comment))]
        else:  # If comment is None, return empty list
            return []

    @property
    def subclass_triples(self) -> list[tuple]:
        if self.sub_class_of:
            return [(self.id_, RDFS.subClassOf, sub_class_of) for sub_class_of in self.sub_class_of]
        else:
            return []

    @property
    def triples(self) -> list[tuple]:
        return self.type_triples + self.label_triples + self.comment_triples + self.subclass_triples


class OWLProperty(OntologyModel):
    id_: URIRef
    type_: set[URIRef]
    label: set[str]
    comment: set[str]
    domain: set[URIRef]
    range_: set[URIRef]
    namespace: Namespace

    @staticmethod
    def same_property_id(definitions: list[InformationProperty]) -> bool:
        return len({definition.property_ for definition in definitions}) == 1

    @classmethod
    def from_list_of_properties(cls, definitions: list[InformationProperty], namespace: Namespace) -> "OWLProperty":
        """Here list of properties is a list of properties with the same id, but different definitions."""

        if not cls.same_property_id(definitions):
            raise exceptions.PropertyDefinitionsNotForSameProperty()

        owl_property = cls.model_construct(
            id_=namespace[definitions[0].property_],
            namespace=namespace,
            label=set(),
            comment=set(),
            domain=set(),
            range_=set(),
            type_=set(),
        )
        for definition in definitions:
            owl_property.type_.add(OWL[definition.type_])
            owl_property.range_.add(
                XSD[definition.value_type.suffix]
                if definition.value_type.suffix in XSD_VALUE_TYPE_MAPPINGS
                else namespace[definition.value_type.suffix]
            )
            owl_property.domain.add(namespace[definition.class_.suffix])
            owl_property.label.add(definition.name or definition.property_)
            if definition.description:
                owl_property.comment.add(definition.description)

        return owl_property

    @field_validator("type_")
    def is_multi_type(cls, v, info: ValidationInfo):
        if len(v) > 1:
            warnings.warn(
                exceptions.OntologyMultiTypeProperty(
                    remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]
                ).message,
                category=exceptions.OntologyMultiTypeProperty,
                stacklevel=2,
            )
        return v

    @field_validator("range_")
    def is_multi_range(cls, v, info: ValidationInfo):
        if len(v) > 1:
            warnings.warn(
                exceptions.OntologyMultiRangeProperty(
                    remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]
                ).message,
                category=exceptions.OntologyMultiRangeProperty,
                stacklevel=2,
            )
        return v

    @field_validator("domain")
    def is_multi_domain(cls, v, info: ValidationInfo):
        if len(v) > 1:
            warnings.warn(
                exceptions.OntologyMultiDomainProperty(
                    remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]
                ).message,
                category=exceptions.OntologyMultiDomainProperty,
                stacklevel=2,
            )
        return v

    @field_validator("label")
    def has_multi_name(cls, v, info: ValidationInfo):
        if len(v) > 1:
            warnings.warn(
                exceptions.OntologyMultiLabeledProperty(remove_namespace(info.data["id_"]), v).message,
                category=exceptions.OntologyMultiLabeledProperty,
                stacklevel=2,
            )
        return v

    @field_validator("comment")
    def has_multi_comment(cls, v, info: ValidationInfo):
        if len(v) > 1:
            warnings.warn(
                exceptions.OntologyMultiDefinitionProperty(remove_namespace(info.data["id_"])).message,
                category=exceptions.OntologyMultiDefinitionProperty,
                stacklevel=2,
            )
        return v

    @property
    def domain_triples(self) -> list[tuple]:
        triples: list[tuple] = []
        if len(self.domain) == 1:
            triples.append((self.id_, RDFS.domain, next(iter(self.domain))))
        else:
            _graph = Graph()
            b_union = BNode()
            b_domain = BNode()
            _graph.add((self.id_, RDFS.domain, b_domain))
            _graph.add((b_domain, OWL.unionOf, b_union))
            _graph.add((b_domain, RDF.type, OWL.Class))
            _ = GraphCollection(_graph, b_union, list(self.domain))
            triples.extend(list(_graph))
        return triples

    @property
    def range_triples(self) -> list[tuple]:
        triples: list[tuple] = []
        if len(self.range_) == 1:
            triples.append((self.id_, RDFS.range, next(iter(self.range_))))
        else:
            _graph = Graph()
            b_union = BNode()
            b_range = BNode()
            _graph.add((self.id_, RDFS.range, b_range))
            _graph.add((b_range, OWL.unionOf, b_union))
            _graph.add((b_range, RDF.type, OWL.Class))
            _graph.add((b_range, OWL.unionOf, b_union))
            _graph.add((b_range, RDF.type, OWL.Class))
            _ = GraphCollection(_graph, b_union, list(self.range_))
            triples.extend(list(_graph))
        return triples

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, type_) for type_ in self.type_]

    @property
    def label_triples(self) -> list[tuple]:
        label = list(filter(None, self.label))
        return [(self.id_, RDFS.label, Literal(label[0] if label else self.id_))]

    @property
    def comment_triples(self) -> list[tuple]:
        return [(self.id_, RDFS.comment, Literal("\n".join(filter(None, self.comment))))]

    @property
    def triples(self) -> list[tuple]:
        return self.type_triples + self.label_triples + self.comment_triples + self.domain_triples + self.range_triples


SHACL = Namespace("http://www.w3.org/ns/shacl#")


class SHACLNodeShape(OntologyModel):
    id_: URIRef
    type_: URIRef = SHACL.NodeShape
    target_class: URIRef
    parent: list[URIRef] | None = None
    property_shapes: list["SHACLPropertyShape"] | None = None
    namespace: Namespace

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, self.type_)]

    @property
    def target_class_triples(self) -> list[tuple]:
        return [(self.id_, SHACL.targetClass, self.target_class)]

    @property
    def target_parent_class_triples(self) -> list[tuple]:
        if self.parent:
            return [(self.id_, RDFS.subClassOf, parent) for parent in self.parent]
        else:
            return []

    @property
    def property_shapes_triples(self) -> list[tuple]:
        triples: list[tuple] = []
        if self.property_shapes:
            for property_shape in self.property_shapes:
                triples.append((self.id_, SHACL.property, property_shape.id_))
                triples.extend(property_shape.triples)
        return triples

    @property
    def triples(self) -> list[tuple]:
        return (
            self.type_triples
            + self.target_class_triples
            + self.property_shapes_triples
            + self.target_parent_class_triples
        )

    @classmethod
    def from_rules(
        cls, class_definition: InformationClass, property_definitions: list[InformationProperty], namespace: Namespace
    ) -> "SHACLNodeShape":
        if class_definition.parent:
            parent = [namespace[parent.suffix + "Shape"] for parent in class_definition.parent]
        else:
            parent = None
        return cls(
            id_=namespace[f"{class_definition.class_.suffix}Shape"],
            target_class=namespace[class_definition.class_.suffix],
            parent=parent,
            property_shapes=[SHACLPropertyShape.from_property(prop, namespace) for prop in property_definitions],
            namespace=namespace,
        )


class SHACLPropertyShape(OntologyModel):
    id_: BNode
    type_: URIRef = SHACL.property
    path: URIRef  # URIRef to property in OWL
    node_kind: URIRef  # SHACL.IRI or SHACL.Literal
    expected_value_type: URIRef
    min_count: int | None
    max_count: int | None
    namespace: Namespace

    @property
    def path_triples(self) -> list[tuple]:
        return [(self.id_, SHACL.path, self.path)]

    @property
    def node_kind_triples(self) -> list[tuple]:
        triples: list[tuple] = [(self.id_, SHACL.nodeKind, self.node_kind)]

        if self.node_kind == SHACL.Literal:
            triples.append((self.id_, SHACL.datatype, self.expected_value_type))
        else:
            triples.append((self.id_, SHACL.node, self.expected_value_type))

        return triples

    @property
    def cardinality_triples(self) -> list[tuple]:
        triples: list[tuple] = []
        if self.min_count:
            triples.append((self.id_, SHACL.minCount, Literal(self.min_count)))
        if self.max_count:
            triples.append((self.id_, SHACL.maxCount, Literal(self.max_count)))

        return triples

    @property
    def triples(self) -> list[tuple]:
        return self.path_triples + self.node_kind_triples + self.cardinality_triples

    @classmethod
    def from_property(cls, definition: InformationProperty, namespace: Namespace) -> "SHACLPropertyShape":
        return cls(
            id_=BNode(),
            path=namespace[definition.property_],
            node_kind=SHACL.IRI if definition.type_ == EntityTypes.object_property else SHACL.Literal,
            expected_value_type=(
                namespace[f"{definition.value_type.suffix}Shape"]
                if definition.type_ == EntityTypes.object_property
                else XSD[definition.value_type.suffix]
            ),
            min_count=definition.min_count,
            max_count=(
                int(definition.max_count) if definition.max_count and definition.max_count != float("inf") else None
            ),
            namespace=namespace,
        )



from __future__ import annotations

from datetime import date, datetime
from typing import ClassVar, cast

from cognite.client.data_classes.data_modeling import (
    Boolean,
    Date,
    FileReference,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    SequenceReference,
    Text,
    TimeSeriesReference,
    Timestamp,
)
from pydantic import BaseModel

from ._base import Entity, EntityTypes


class ValueTypeMapping(BaseModel):
    """Mapping between XSD, Python, DMS and Graphql types."""

    xsd: str
    python: type
    dms: type
    graphql: str
    sql: str


# mypy: ignore-errors
class XSDValueType(Entity):
    """Value type is a data/object type defined as a child of Entity model."""

    type_: ClassVar[EntityTypes] = EntityTypes.xsd_value_type
    mapping: ValueTypeMapping

    @property
    def python(self) -> type:
        """Returns the Python type for a given value type."""
        return cast(ValueTypeMapping, self.mapping).python

    @property
    def xsd(self) -> str:
        """Returns the XSD type for a given value type."""
        return cast(ValueTypeMapping, self.mapping).xsd

    @property
    def dms(self) -> type:
        """Returns the DMS type for a given value type."""
        return cast(ValueTypeMapping, self.mapping).dms

    @property
    def graphql(self) -> str:
        """Returns the Graphql type for a given value type."""
        return cast(ValueTypeMapping, self.mapping).graphql

    @property
    def sql(self) -> str:
        """Returns the SQL type for a given value type."""
        return cast(ValueTypeMapping, self.mapping).sql


class DMSValueType(XSDValueType):
    type_: ClassVar[EntityTypes] = EntityTypes.dms_value_type

    def __str__(self) -> str:
        return self.dms._type


_DATA_TYPES: list[dict] = [
    {"name": "boolean", "python": bool, "GraphQL": "Boolean", "dms": Boolean, "SQL": "BOOLEAN"},
    {"name": "float", "python": float, "GraphQL": "Float", "dms": Float64, "SQL": "FLOAT"},
    {"name": "double", "python": float, "GraphQL": "Float", "dms": Float64, "SQL": "DOUBLE"},
    {"name": "integer", "python": int, "GraphQL": "Int", "dms": Int32, "SQL": "INTEGER"},
    {"name": "nonPositiveInteger", "python": int, "GraphQL": "Int", "dms": Int32, "SQL": "INTEGER"},
    {"name": "nonNegativeInteger", "python": int, "GraphQL": "Int", "dms": Int32, "SQL": "INTEGER"},
    {"name": "negativeInteger", "python": int, "GraphQL": "Int", "dms": Int32, "SQL": "INTEGER"},
    {"name": "long", "python": int, "GraphQL": "Int", "dms": Int64, "SQL": "INTEGER"},
    {"name": "string", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "langString", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "anyURI", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "normalizedString", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "token", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    # Graphql does not have a datetime/date type this is CDF specific
    {"name": "dateTime", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp, "SQL": "TIMESTAMP"},
    {"name": "dateTimeStamp", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp, "SQL": "TIMESTAMP"},
    {"name": "date", "python": date, "GraphQL": "String", "dms": Date, "SQL": "DATE"},
    # Some RDF types which are not in XSD
    {"name": "PlainLiteral", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "Literal", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    # CDF specific types, not in XSD
    {
        "name": "timeseries",
        "python": TimeSeriesReference,
        "GraphQL": "TimeSeries",
        "dms": TimeSeriesReference,
        "SQL": "STRING",
    },
    {"name": "file", "python": FileReference, "GraphQL": "File", "dms": FileReference, "SQL": "STRING"},
    {"name": "sequence", "python": SequenceReference, "GraphQL": "Sequence", "dms": SequenceReference, "SQL": "STRING"},
    {"name": "json", "python": Json, "GraphQL": "Json", "dms": Json, "SQL": "STRING"},
]

_DMS_TYPES: list[dict] = [
    {"name": "boolean", "python": bool, "GraphQL": "Boolean", "dms": Boolean, "SQL": "BOOLEAN"},
    {"name": "float", "python": float, "GraphQL": "Float", "dms": Float32, "SQL": "FLOAT"},
    {"name": "double", "python": float, "GraphQL": "Float", "dms": Float64, "SQL": "DOUBLE"},
    {"name": "integer", "python": int, "GraphQL": "Int", "dms": Int32, "SQL": "INTEGER"},
    {"name": "long", "python": int, "GraphQL": "Int", "dms": Int64, "SQL": "INTEGER"},
    {"name": "string", "python": str, "GraphQL": "String", "dms": Text, "SQL": "STRING"},
    {"name": "dateTimeStamp", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp, "SQL": "TIMESTAMP"},
    {
        "name": "timeseries",
        "python": TimeSeriesReference,
        "GraphQL": "TimeSeries",
        "dms": TimeSeriesReference,
        "SQL": "STRING",
    },
    {"name": "file", "python": FileReference, "GraphQL": "File", "dms": FileReference, "SQL": "STRING"},
    {"name": "sequence", "python": SequenceReference, "GraphQL": "Sequence", "dms": SequenceReference, "SQL": "STRING"},
    {"name": "json", "python": Json, "GraphQL": "Json", "dms": Json, "SQL": "STRING"},
]

XSD_VALUE_TYPE_MAPPINGS: dict[str, XSDValueType] = {
    data_type["name"]: XSDValueType(
        prefix="xsd",
        suffix=cast(str, data_type["name"]),
        name=cast(str, data_type["name"]),
        mapping=ValueTypeMapping(
            xsd=data_type["name"],
            python=data_type["python"],
            dms=data_type["dms"],
            graphql=data_type["GraphQL"],
            sql=data_type["SQL"],
        ),
    )
    for data_type in _DATA_TYPES
}

DMS_VALUE_TYPE_MAPPINGS: dict[str, DMSValueType] = {
    data_type["dms"]._type.casefold(): DMSValueType(
        prefix="dms",
        suffix=data_type["dms"]._type.casefold(),
        name=data_type["dms"]._type.casefold(),
        mapping=ValueTypeMapping(
            xsd=data_type["name"],
            python=data_type["python"],
            dms=data_type["dms"],
            graphql=data_type["GraphQL"],
            sql=data_type["SQL"],
        ),
    )
    for data_type in _DMS_TYPES
}



import json
import sys
import warnings
import zipfile
from collections import Counter, defaultdict
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, ClassVar, cast

import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import DatabaseWrite, DatabaseWriteList, TransformationWrite, TransformationWriteList
from cognite.client.data_classes.data_modeling import ViewApply
from cognite.client.data_classes.transformations.common import Edges, EdgeType, Nodes, ViewInfo

from cognite.neat.rules.issues.dms import (
    ContainerPropertyUsedMultipleTimesError,
    DirectRelationMissingSourceWarning,
    DMSSchemaError,
    DuplicatedViewInDataModelError,
    MissingContainerError,
    MissingContainerPropertyError,
    MissingEdgeViewError,
    MissingParentViewError,
    MissingSourceViewError,
    MissingSpaceError,
    MissingViewError,
)
from cognite.neat.rules.models._rules._types._value import DMS_VALUE_TYPE_MAPPINGS
from cognite.neat.utils.cdf_loaders import ViewLoader
from cognite.neat.utils.cdf_loaders.data_classes import RawTableWrite, RawTableWriteList
from cognite.neat.utils.text import to_camel

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class DMSSchema:
    spaces: dm.SpaceApplyList = field(default_factory=lambda: dm.SpaceApplyList([]))
    data_models: dm.DataModelApplyList = field(default_factory=lambda: dm.DataModelApplyList([]))
    views: dm.ViewApplyList = field(default_factory=lambda: dm.ViewApplyList([]))
    containers: dm.ContainerApplyList = field(default_factory=lambda: dm.ContainerApplyList([]))
    node_types: dm.NodeApplyList = field(default_factory=lambda: dm.NodeApplyList([]))
    # The frozen ids are parts of the schema that should not be modified or deleted.
    # This is used the exporting the schema.
    frozen_ids: set[dm.ViewId | dm.ContainerId | dm.NodeId] = field(default_factory=set)

    _FIELD_NAME_BY_RESOURCE_TYPE: ClassVar[dict[str, str]] = {
        "container": "containers",
        "view": "views",
        "datamodel": "data_models",
        "space": "spaces",
        "node": "node_types",
    }

    @classmethod
    def from_model_id(cls, client: CogniteClient, data_model_id: dm.DataModelIdentifier) -> "DMSSchema":
        data_models = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        if len(data_models) == 0:
            raise ValueError(f"Data model {data_model_id} not found")
        data_model = data_models.latest_version()
        return cls.from_data_model(client, data_model)

    @classmethod
    def from_data_model(cls, client: CogniteClient, data_model: dm.DataModel) -> "DMSSchema":
        views = dm.ViewList(data_model.views)
        container_ids = views.referenced_containers()
        containers = client.data_modeling.containers.retrieve(list(container_ids))
        cls._append_referenced_containers(client, containers)

        space_read = client.data_modeling.spaces.retrieve(data_model.space)
        if space_read is None:
            raise ValueError(f"Space {data_model.space} not found")
        space = space_read.as_write()
        data_model_write = data_model.as_write()
        data_model_write.views = list(views.as_write())

        view_loader = ViewLoader(client)
        # We need to include parent views in the schema to make sure that the schema is valid.
        existing_view_ids = set(views.as_ids())
        parent_view_ids = {parent for view in views for parent in view.implements or []}
        parents = view_loader.retrieve_all_parents(list(parent_view_ids - existing_view_ids))
        views.extend([parent for parent in parents if parent.as_id() not in existing_view_ids])

        # Converting views from read to write format requires to account for parents (implements)
        # as the read format contains all properties from all parents, while the write formate should not contain
        # properties from any parents.
        # The ViewLoader as_write method looks up parents and remove properties from them.
        view_write = dm.ViewApplyList([view_loader.as_write(view) for view in views])

        return cls(
            spaces=dm.SpaceApplyList([space]),
            data_models=dm.DataModelApplyList([data_model_write]),
            views=view_write,
            containers=containers.as_write(),
        )

    @classmethod
    def from_directory(cls, directory: str | Path) -> Self:
        """Load a schema from a directory containing YAML files.

        The directory is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data = cls._read_directory(Path(directory))
        return cls.load(data)

    @classmethod
    def _read_directory(cls, directory: Path) -> dict[str, list[Any]]:
        data: dict[str, Any] = {}
        for yaml_file in directory.rglob("*.yaml"):
            if "." in yaml_file.stem:
                resource_type = yaml_file.stem.rsplit(".", 1)[-1]
                if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type):
                    data.setdefault(attr_name, [])
                    loaded = yaml.safe_load(yaml_file.read_text())
                    if isinstance(loaded, list):
                        data[attr_name].extend(loaded)
                    else:
                        data[attr_name].append(loaded)
        return data

    def to_directory(
        self,
        directory: str | Path,
        exclude: set[str] | None = None,
        new_line: str | None = "\n",
        encoding: str | None = "utf-8",
    ) -> None:
        """Save the schema to a directory as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            directory (str | Path): The directory to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
            new_line (str): The line endings to use in the output files. Defaults to "\n".
            encoding (str): The encoding to use in the output files. Defaults to "utf-8".
        """
        path_dir = Path(directory)
        exclude_set = exclude or set()
        data_models = path_dir / "data_models"
        data_models.mkdir(parents=True, exist_ok=True)
        if "spaces" not in exclude_set:
            for space in self.spaces:
                (data_models / f"{space.space}.space.yaml").write_text(
                    space.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "data_models" not in exclude_set:
            for model in self.data_models:
                (data_models / f"{model.external_id}.datamodel.yaml").write_text(
                    model.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "views" not in exclude_set and self.views:
            view_dir = data_models / "views"
            view_dir.mkdir(parents=True, exist_ok=True)
            for view in self.views:
                (view_dir / f"{view.external_id}.view.yaml").write_text(
                    view.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "containers" not in exclude_set and self.containers:
            container_dir = data_models / "containers"
            container_dir.mkdir(parents=True, exist_ok=True)
            for container in self.containers:
                (container_dir / f"{container.external_id}.container.yaml").write_text(
                    container.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "node_types" not in exclude_set and self.node_types:
            node_dir = data_models / "nodes"
            node_dir.mkdir(parents=True, exist_ok=True)
            for node in self.node_types:
                (node_dir / f"{node.external_id}.node.yaml").write_text(
                    node.dump_yaml(), newline=new_line, encoding=encoding
                )

    @classmethod
    def from_zip(cls, zip_file: str | Path) -> Self:
        """Load a schema from a ZIP file containing YAML files.

        The ZIP file is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data = cls._read_zip(Path(zip_file))
        return cls.load(data)

    @classmethod
    def _read_zip(cls, zip_file: Path) -> dict[str, list[Any]]:
        data: dict[str, list[Any]] = {}
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename.endswith(".yaml"):
                    if "/" not in file_info.filename:
                        continue
                    filename = Path(file_info.filename.split("/")[-1])
                    if "." not in filename.stem:
                        continue
                    resource_type = filename.stem.rsplit(".", 1)[-1]
                    if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type):
                        data.setdefault(attr_name, [])
                        loaded = yaml.safe_load(zip_ref.read(file_info).decode())
                        if isinstance(loaded, list):
                            data[attr_name].extend(loaded)
                        else:
                            data[attr_name].append(loaded)
        return data

    def to_zip(self, zip_file: str | Path, exclude: set[str] | None = None) -> None:
        """Save the schema to a ZIP file as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            zip_file (str | Path): The ZIP file to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
        """
        exclude_set = exclude or set()
        with zipfile.ZipFile(zip_file, "w") as zip_ref:
            if "spaces" not in exclude_set:
                for space in self.spaces:
                    zip_ref.writestr(f"data_models/{space.space}.space.yaml", space.dump_yaml())
            if "data_models" not in exclude_set:
                for model in self.data_models:
                    zip_ref.writestr(f"data_models/{model.external_id}.datamodel.yaml", model.dump_yaml())
            if "views" not in exclude_set:
                for view in self.views:
                    zip_ref.writestr(f"data_models/views/{view.external_id}.view.yaml", view.dump_yaml())
            if "containers" not in exclude_set:
                for container in self.containers:
                    zip_ref.writestr(
                        f"data_models/containers{container.external_id}.container.yaml", container.dump_yaml()
                    )
            if "node_types" not in exclude_set:
                for node in self.node_types:
                    zip_ref.writestr(f"data_models/nodes/{node.external_id}.node.yaml", node.dump_yaml())

    @classmethod
    def load(cls, data: str | dict[str, Any]) -> Self:
        if isinstance(data, str):
            # YAML is a superset of JSON, so we can use the same parser
            data_dict = yaml.safe_load(data)
        else:
            data_dict = data
        loaded: dict[str, Any] = {}
        for attr in fields(cls):
            if items := data_dict.get(attr.name) or data_dict.get(to_camel(attr.name)):
                loaded[attr.name] = attr.type.load(items)
        return cls(**loaded)

    def dump(self, camel_case: bool = True, sort: bool = True) -> dict[str, Any]:
        """Dump the schema to a dictionary that can be serialized to JSON.

        Args:
            camel_case (bool): If True, the keys in the output dictionary will be in camel case.
            sort (bool): If True, the items in the output dictionary will be sorted by their ID.
                This is useful for deterministic output which is useful for comparing schemas.

        Returns:
            dict: The schema as a dictionary.
        """
        output: dict[str, Any] = {}
        cls_fields = sorted(fields(self), key=lambda f: f.name) if sort else fields(self)
        for attr in cls_fields:
            if items := getattr(self, attr.name):
                items = sorted(items, key=self._to_sortable_identifier) if sort else items
                key = to_camel(attr.name) if camel_case else attr.name
                output[key] = [item.dump(camel_case=camel_case) for item in items]
        return output

    @classmethod
    def _to_sortable_identifier(cls, item: Any) -> str | tuple[str, str] | tuple[str, str, str]:
        if isinstance(item, dm.ContainerApply | dm.ViewApply | dm.DataModelApply | dm.NodeApply | RawTableWrite):
            identifier = item.as_id().as_tuple()
            if len(identifier) == 3 and identifier[2] is None:
                return identifier[:2]  # type: ignore[misc]
            return cast(tuple[str, str] | tuple[str, str, str], identifier)
        elif isinstance(item, dm.SpaceApply):
            return item.space
        elif isinstance(item, TransformationWrite):
            return item.external_id or ""
        elif isinstance(item, DatabaseWrite):
            return item.name or ""
        else:
            raise ValueError(f"Cannot sort item of type {type(item)}")

    def validate(self) -> list[DMSSchemaError]:
        errors: set[DMSSchemaError] = set()
        defined_spaces = {space.space for space in self.spaces}
        defined_containers = {container.as_id(): container for container in self.containers}
        defined_views = {view.as_id() for view in self.views}

        for container in self.containers:
            if container.space not in defined_spaces:
                errors.add(MissingSpaceError(space=container.space, referred_by=container.as_id()))

        for view in self.views:
            view_id = view.as_id()
            if view.space not in defined_spaces:
                errors.add(MissingSpaceError(space=view.space, referred_by=view_id))

            for parent in view.implements or []:
                if parent not in defined_views:
                    errors.add(MissingParentViewError(view=parent, referred_by=view_id))

            for prop_name, prop in (view.properties or {}).items():
                if isinstance(prop, dm.MappedPropertyApply):
                    ref_container = defined_containers.get(prop.container)
                    if ref_container is None:
                        errors.add(MissingContainerError(container=prop.container, referred_by=view_id))
                    elif prop.container_property_identifier not in ref_container.properties:
                        errors.add(
                            MissingContainerPropertyError(
                                container=prop.container,
                                property=prop.container_property_identifier,
                                referred_by=view_id,
                            )
                        )
                    else:
                        container_property = ref_container.properties[prop.container_property_identifier]

                        if isinstance(container_property.type, dm.DirectRelation) and prop.source is None:
                            warnings.warn(
                                DirectRelationMissingSourceWarning(view_id=view_id, property=prop_name), stacklevel=2
                            )

                if isinstance(prop, dm.EdgeConnectionApply) and prop.source not in defined_views:
                    errors.add(MissingSourceViewError(view=prop.source, property=prop_name, referred_by=view_id))

                if (
                    isinstance(prop, dm.EdgeConnectionApply)
                    and prop.edge_source is not None
                    and prop.edge_source not in defined_views
                ):
                    errors.add(MissingEdgeViewError(view=prop.edge_source, property=prop_name, referred_by=view_id))

            property_count = Counter(
                (prop.container, prop.container_property_identifier)
                for prop in (view.properties or {}).values()
                if isinstance(prop, dm.MappedPropertyApply)
            )
            for (container_id, container_property_identifier), count in property_count.items():
                if count > 1:
                    view_properties = [
                        prop_name
                        for prop_name, prop in (view.properties or {}).items()
                        if isinstance(prop, dm.MappedPropertyApply)
                        and (prop.container, prop.container_property_identifier)
                        == (container_id, container_property_identifier)
                    ]
                    errors.add(
                        ContainerPropertyUsedMultipleTimesError(
                            container=container_id,
                            property=container_property_identifier,
                            referred_by=frozenset({(view_id, prop_name) for prop_name in view_properties}),
                        )
                    )

        for model in self.data_models:
            if model.space not in defined_spaces:
                errors.add(MissingSpaceError(space=model.space, referred_by=model.as_id()))

            view_counts: dict[dm.ViewId, int] = defaultdict(int)
            for view_id_or_class in model.views or []:
                view_id = view_id_or_class if isinstance(view_id_or_class, dm.ViewId) else view_id_or_class.as_id()
                if view_id not in defined_views:
                    errors.add(MissingViewError(referred_by=model.as_id(), view=view_id))
                view_counts[view_id] += 1

            for view_id, count in view_counts.items():
                if count > 1:
                    errors.add(DuplicatedViewInDataModelError(referred_by=model.as_id(), view=view_id))

        return list(errors)

    @classmethod
    def _append_referenced_containers(cls, client: CogniteClient, containers: dm.ContainerList) -> None:
        """Containers can reference each other through the 'requires' constraint.

        This method retrieves all containers that are referenced by other containers through the 'requires' constraint,
        including their parents.

        """
        for _ in range(10):  # Limiting the number of iterations to avoid infinite loops
            referenced_containers = {
                const.require
                for container in containers
                for const in (container.constraints or {}).values()
                if isinstance(const, dm.RequiresConstraint)
            }
            missing_containers = referenced_containers - set(containers.as_ids())
            if not missing_containers:
                break
            found_containers = client.data_modeling.containers.retrieve(list(missing_containers))
            containers.extend(found_containers)
            if len(found_containers) != len(missing_containers):
                break
        else:
            warnings.warn(
                "The maximum number of iterations was reached while resolving referenced containers."
                "There might be referenced containers that are not included in the list of containers.",
                RuntimeWarning,
                stacklevel=2,
            )
        return None


@dataclass
class PipelineSchema(DMSSchema):
    transformations: TransformationWriteList = field(default_factory=lambda: TransformationWriteList([]))
    databases: DatabaseWriteList = field(default_factory=lambda: DatabaseWriteList([]))
    raw_tables: RawTableWriteList = field(default_factory=lambda: RawTableWriteList([]))

    _FIELD_NAME_BY_RESOURCE_TYPE: ClassVar[dict[str, str]] = {
        **DMSSchema._FIELD_NAME_BY_RESOURCE_TYPE,
        "raw": "raw_tables",
    }

    def __post_init__(self):
        existing_databases = {database.name for database in self.databases}
        table_database = {table.database for table in self.raw_tables}
        if missing := table_database - existing_databases:
            self.databases.extend([DatabaseWrite(name=database) for database in missing])

    @classmethod
    def _read_directory(cls, directory: Path) -> dict[str, list[Any]]:
        data = super()._read_directory(directory)
        for yaml_file in directory.rglob("*.yaml"):
            if yaml_file.parent.name in ("transformations", "raw"):
                attr_name = cls._FIELD_NAME_BY_RESOURCE_TYPE.get(yaml_file.parent.name, yaml_file.parent.name)
                data.setdefault(attr_name, [])
                loaded = yaml.safe_load(yaml_file.read_text())
                if isinstance(loaded, list):
                    data[attr_name].extend(loaded)
                else:
                    data[attr_name].append(loaded)
        return data

    def to_directory(
        self,
        directory: str | Path,
        exclude: set[str] | None = None,
        new_line: str | None = "\n",
        encoding: str | None = "utf-8",
    ) -> None:
        super().to_directory(directory, exclude)
        exclude_set = exclude or set()
        path_dir = Path(directory)
        if "transformations" not in exclude_set and self.transformations:
            transformation_dir = path_dir / "transformations"
            transformation_dir.mkdir(exist_ok=True, parents=True)
            for transformation in self.transformations:
                (transformation_dir / f"{transformation.external_id}.yaml").write_text(
                    transformation.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "raw" not in exclude_set and self.raw_tables:
            # The RAW Databases are not written to file. This is because cognite-toolkit expects the RAW databases
            # to be in the same file as the RAW tables.
            raw_dir = path_dir / "raw"
            raw_dir.mkdir(exist_ok=True, parents=True)
            for raw_table in self.raw_tables:
                (raw_dir / f"{raw_table.name}.yaml").write_text(
                    raw_table.dump_yaml(), newline=new_line, encoding=encoding
                )

    def to_zip(self, zip_file: str | Path, exclude: set[str] | None = None) -> None:
        super().to_zip(zip_file, exclude)
        exclude_set = exclude or set()
        with zipfile.ZipFile(zip_file, "a") as zip_ref:
            if "transformations" not in exclude_set:
                for transformation in self.transformations:
                    zip_ref.writestr(f"transformations/{transformation.external_id}.yaml", transformation.dump_yaml())
            if "raw" not in exclude_set:
                # The RAW Databases are not written to file. This is because cognite-toolkit expects the RAW databases
                # to be in the same file as the RAW tables.
                for raw_table in self.raw_tables:
                    zip_ref.writestr(f"raw/{raw_table.name}.yaml", raw_table.dump_yaml())

    @classmethod
    def _read_zip(cls, zip_file: Path) -> dict[str, list[Any]]:
        data = super()._read_zip(zip_file)
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename.endswith(".yaml"):
                    if "/" not in file_info.filename:
                        continue
                    filepath = Path(file_info.filename)
                    if (parent := filepath.parent.name) in ("transformations", "raw"):
                        attr_name = cls._FIELD_NAME_BY_RESOURCE_TYPE.get(parent, parent)
                        data.setdefault(attr_name, [])
                        loaded = yaml.safe_load(zip_ref.read(file_info).decode())
                        if isinstance(loaded, list):
                            data[attr_name].extend(loaded)
                        else:
                            data[attr_name].append(loaded)
        return data

    @classmethod
    def from_dms(cls, schema: DMSSchema, instance_space: str | None = None) -> "PipelineSchema":
        if not schema.data_models:
            raise ValueError("PipelineSchema must contain at least one data model")
        first_data_model = schema.data_models[0]
        # The database name is limited to 32 characters
        database_name = first_data_model.external_id[:32]
        instance_space = instance_space or first_data_model.space
        database = DatabaseWrite(name=database_name)
        parent_views = {parent for view in schema.views for parent in view.implements or []}
        container_by_id = {container.as_id(): container for container in schema.containers}

        transformations = TransformationWriteList([])
        raw_tables = RawTableWriteList([])
        for view in schema.views:
            if view.as_id() in parent_views:
                # Skipping parents as they do not have their own data
                continue
            mapped_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.MappedPropertyApply)
            }
            if mapped_properties:
                view_table = RawTableWrite(name=f"{view.external_id}Properties", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_property_transformation(
                    mapped_properties, view, view_table, container_by_id, instance_space
                )
                transformations.append(transformation)
            connection_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.EdgeConnectionApply)
            }
            for prop_name, connection_property in connection_properties.items():
                view_table = RawTableWrite(name=f"{view.external_id}.{prop_name}Edge", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_edge_transformation(connection_property, view, view_table, instance_space)
                transformations.append(transformation)

        return cls(
            spaces=schema.spaces,
            data_models=schema.data_models,
            views=schema.views,
            containers=schema.containers,
            transformations=transformations,
            databases=DatabaseWriteList([database]),
            raw_tables=raw_tables,
        )

    @classmethod
    def _create_property_transformation(
        cls,
        properties: dict[str, dm.MappedPropertyApply],
        view: ViewApply,
        table: RawTableWrite,
        container_by_id: dict[dm.ContainerId, dm.ContainerApply],
        instance_space: str,
    ) -> TransformationWrite:
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            # 'mappings' is set here and overwritten further down to ensure the correct order
            "mappings": [],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        mappings = [
            {"from": "externalId", "to": "externalId", "asType": "STRING"},
        ]
        select_rows = ["cast(`externalId` as STRING) as externalId"]
        for prop_name, prop in properties.items():
            container = container_by_id.get(prop.container)
            if container is not None:
                dms_type = container.properties[prop.container_property_identifier].type._type
                if dms_type in DMS_VALUE_TYPE_MAPPINGS:
                    sql_type = DMS_VALUE_TYPE_MAPPINGS[dms_type].sql
                else:
                    warnings.warn(
                        f"Unknown DMS type '{dms_type}' for property '{prop_name}'", RuntimeWarning, stacklevel=2
                    )
                    sql_type = "STRING"
            else:
                sql_type = "STRING"
            select_rows.append(f"cast(`{prop_name}` as {sql_type}) as {prop_name}")
            mappings.append({"from": prop_name, "to": prop_name, "asType": sql_type})
        mapping_mode["mappings"] = mappings
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Nodes(
                view=ViewInfo(view.space, view.external_id, view.version),
                instance_space=instance_space,
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )

    @classmethod
    def _create_edge_transformation(
        cls, property_: dm.EdgeConnectionApply, view: ViewApply, table: RawTableWrite, instance_space: str
    ) -> TransformationWrite:
        start, end = view.external_id, property_.source.external_id
        if property_.direction == "inwards":
            start, end = end, start
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            "mappings": [
                {"from": "externalId", "to": "externalId", "asType": "STRING"},
                {"from": start, "to": "startNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
                {"from": end, "to": "endNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
            ],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        select_rows = [
            "cast(`externalId` as STRING) as externalId",
            f"node_reference('{instance_space}', `{start}`) as startNode",
            f"node_reference('{instance_space}', `{end}`) as endNode",
        ]
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Edges(
                instance_space=instance_space,
                edge_type=EdgeType(space=property_.type.space, external_id=property_.type.external_id),
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )



import base64
import time
from pathlib import Path
from typing import ClassVar

import requests
from cognite.client import CogniteClient

from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.step_model import Configurable, Step

CATEGORY = "IO Steps"


__all__ = [
    "DownloadFileFromGitHub",
    "UploadFileToGitHub",
    "DownloadFileFromCDF",
    "UploadFileToCDF",
    "DownloadDataFromRestApiToFile",
]


class DownloadFileFromGitHub(Step):
    """
    This step fetches and stores the file from private Github repository
    """

    description = "This step fetches and stores the file from private Github repository"
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="github.filepath", value="", label="File path to the file stored on Github"),
        Configurable(
            name="github.personal_token",
            value="",
            label="Github Personal Access Token which allows fetching file from private Github repository",
            type="password",
        ),
        Configurable(name="github.owner", value="", label="Github repository owner, also know as github organization"),
        Configurable(name="github.repo", value="", label="Github repository from which the file is being fetched"),
        Configurable(
            name="github.branch", value="main", label="Github repository branch from which the file is being fetched"
        ),
        Configurable(
            name="local.file_name", value="", label="The name of the file under which it will be stored locally"
        ),
        Configurable(name="local.storage_dir", value="rules/", label="The directory where the file will be stored"),
    ]

    def run(self) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        github_filepath = self.configs["github.filepath"]
        github_personal_token = self.configs["github.personal_token"]
        github_owner = self.configs["github.owner"]
        github_repo = self.configs["github.repo"]
        github_branch = self.configs["github.branch"]
        github_file_name = Path(github_filepath).name
        local_file_name = self.configs["local.file_name"] or github_file_name
        full_local_file_path = self.data_store_path / Path(self.configs["local.storage_dir"])

        if not full_local_file_path.exists():
            full_local_file_path.mkdir(parents=True, exist_ok=True)

        r = requests.get(
            f"https://api.github.com/repos/{github_owner}/{github_repo }"
            + f"/contents/{github_filepath}?ref={github_branch}",
            headers={"accept": "application/vnd.github.v3.raw", "authorization": f"token {github_personal_token}"},
        )

        if r.status_code >= 200 and r.status_code < 300:
            local_download_path = Path(self.configs["local.storage_dir"]) / local_file_name
            full_local_file_path = full_local_file_path / local_file_name
            try:
                with full_local_file_path.open("wb") as f:
                    f.write(r.content)
            except Exception as e:
                return FlowMessage(
                    error_text=f"Error writing file to {full_local_file_path}. Error: {e}",
                    step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
                )
        else:
            return FlowMessage(
                error_text=f"Error fetching file from Github: {r.text}",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        output_text = (
            "<p></p>"
            f" Downloaded file <b>{github_file_name}</b> from:"
            f'<p><a href="https://github.com/{github_owner}/{github_repo}/tree/{github_branch}"'
            f'target="_blank">https://github.com/{github_owner}/{github_repo}/tree/{github_branch}</a></p>'
        )

        output_text += (
            "<p></p>"
            " Downloaded rules accessible locally under file name "
            f'<a href="/data/{local_download_path}?{time.time()}" '
            f'target="_blank">{local_file_name}</a>'
        )

        return FlowMessage(output_text=output_text)


class UploadFileToGitHub(Step):
    """
    This step uploads file to private Github repository
    """

    description = "This step uploads file to private Github repository"
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="github.filepath", value="", label="File path to the file stored on Github"),
        Configurable(
            name="github.personal_token",
            value="",
            label="Github Personal Access Token which allows uploading file to private Github repository",
            type="password",
        ),
        Configurable(name="github.owner", value="", label="Github repository owner, also know as github organization"),
        Configurable(name="github.repo", value="", label="Github repository the file is being uploaded to"),
        Configurable(
            name="github.branch", value="main", label="Github repository branch the file is being uploaded to"
        ),
        Configurable(
            name="github.commit_message",
            value="New file",
            label="The commit message to be used when uploading the file",
        ),
        Configurable(name="local.file_name", value="", label="The name of the local file to be uploaded to Github"),
        Configurable(name="local.storage_dir", value="rules/", label="Local directory where the file is stored"),
    ]

    def run(self) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        github_filepath = self.configs["github.filepath"]
        github_personal_token = self.configs["github.personal_token"]
        github_owner = self.configs["github.owner"]
        github_repo = self.configs["github.repo"]
        github_branch = self.configs["github.branch"]
        local_file_name = self.configs["local.file_name"]
        full_local_file_path = self.data_store_path / Path(self.configs["local.storage_dir"]) / local_file_name

        if not full_local_file_path.exists():
            return FlowMessage(
                error_text=f"File {full_local_file_path} doesn't exist",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        with full_local_file_path.open("rb") as f:
            file_content = f.read()

        headers = {"Authorization": f"Bearer {github_personal_token}", "Content-Type": "application/json"}

        # Create a content object
        content = {
            "message": self.configs["github.commit_message"],
            "content": base64.b64encode(file_content).decode("utf-8"),
            "branch": github_branch,
        }
        base_url = f"https://api.github.com/repos/{github_owner}/{github_repo }/contents/{github_filepath}"

        # Check if the file exists already
        response = requests.get(f"{base_url}?ref={github_branch}", headers=headers)

        if response.status_code == 200:
            # File exists, update it
            sha = response.json()["sha"]
            content["sha"] = sha
            response = requests.put(base_url, headers=headers, json=content)
        elif response.status_code == 404:
            # File doesn't exist, create it
            response = requests.put(base_url, headers=headers, json=content)
        else:
            # Unexpected response
            return FlowMessage(
                error_text=f"Unexpected response from Github: {response.text}",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )
        if response.status_code == 200 or response.status_code == 201:
            return FlowMessage(output_text=f"File {local_file_name} uploaded to Github successfully")
        else:
            return FlowMessage(
                error_text=f"Error uploading file to Github: {response.text}",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )


class DownloadFileFromCDF(Step):
    """
    This step fetches and stores file from CDF
    """

    description = "This step fetches and stores file from CDF"
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="cdf.external_id", value="", label="External ID of the file stored in CDF"),
        Configurable(
            name="local.file_name",
            value="",
            label="The name of the file under which the content will be stored locally",
        ),
        Configurable(name="local.storage_dir", value="rules/", label="The directory where the file will be stored"),
    ]

    def run(self, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        output_dir = self.data_store_path / Path(self.configs["local.storage_dir"])
        output_dir.mkdir(parents=True, exist_ok=True)

        full_local_file_path = output_dir / self.configs["local.file_name"]
        cdf_client.files.download_to_path(full_local_file_path, external_id=self.configs["cdf.external_id"])
        if full_local_file_path.exists():
            return FlowMessage(output_text=f"File {self.configs['local.file_name']} downloaded from CDF successfully")
        else:
            return FlowMessage(
                error_text="Error downloading file from CDF", step_execution_status=StepExecutionStatus.ABORT_AND_FAIL
            )


class UploadFileToCDF(Step):
    """
    This step uploads file to CDF
    """

    description = "This step uploads file to CDF"
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="cdf.external_id", value="", label="Exernal Id for the file to be stored in CDF"),
        Configurable(
            name="cdf.dataset_id", value="", label="Dataset Id for the file to be stored in CDF. Must be a number"
        ),
        Configurable(name="local.file_name", value="", label="The name of the local file to be uploaded to CDF"),
        Configurable(name="local.storage_dir", value="rules/", label="Local directory where the file is stored"),
    ]

    def run(self, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        full_local_file_path = (
            self.data_store_path / Path(self.configs["local.storage_dir"]) / self.configs["local.file_name"]
        )
        dataset_id = int(self.configs["cdf.dataset_id"]) if self.configs["cdf.dataset_id"].isdigit() else None
        cdf_client.files.upload(
            str(full_local_file_path),
            external_id=self.configs["cdf.external_id"],
            overwrite=True,
            data_set_id=dataset_id,
        )
        return FlowMessage(output_text=f"File {self.configs['local.file_name']} uploaded to CDF successfully")


class DownloadDataFromRestApiToFile(Step):
    """
    This step downloads the response from a REST API and saves it to a file.
    """

    description = "This step downloads the response from a REST API and saves it to a file."
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="api_url",
            value="",
            label="API URL",
        ),
        Configurable(
            name="output_file_path",
            value="workflows/workflow_name/output.json",
            label="Output File Path. The path must be relative to the data store path.",
        ),
        Configurable(
            name="http_method",
            value="GET",
            label="HTTP Method (GET/POST/PUT)",
            options=["GET", "POST", "PUT"],
        ),
        Configurable(
            name="auth_mode",
            value="none",
            label="Authentication Mode (basic/token/none)",
            options=["basic", "token", "none"],
        ),
        Configurable(
            name="username",
            value="",
            label="Username (for basic auth)",
        ),
        Configurable(
            name="password",
            value="",
            label="Password (for basic auth)",
            type="password",
        ),
        Configurable(
            name="token",
            value="",
            label="Token (for token auth)",
            type="password",
        ),
        Configurable(
            name="response_destination",
            value="file",
            label="Destination for the response (file/flow_message/both)",
            options=["file", "flow_message", "both"],
        ),
        Configurable(
            name="http_headers",
            value="",
            label="Custom HTTP headers separated by ';' . Example: \
              'Content-Type: application/json; Accept: application/json'",
        ),
    ]

    def run(self) -> FlowMessage:  # type: ignore[override, syntax]
        api_url = self.configs["api_url"]
        output_file_path = Path(self.data_store_path) / Path(self.configs["output_file_path"])

        output_file_path.parent.mkdir(parents=True, exist_ok=True)

        http_method = self.configs["http_method"].upper()
        auth_mode = self.configs["auth_mode"]
        username = self.configs["username"]
        password = self.configs["password"]
        token = self.configs["token"]
        http_headers_str = self.configs.get("http_headers", "")
        headers = {}
        for header in http_headers_str.split(";"):
            if header:
                key, value = header.split(":")
                headers[key.strip()] = value.strip()
        try:
            if auth_mode == "basic":
                if username and password:
                    headers["Authorization"] = f'Basic {base64.b64encode(f"{username}:{password}".encode()).decode()}'
                else:
                    return FlowMessage(
                        error_text="Username and password are required for Basic Authentication",
                        step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
                    )
            elif auth_mode == "token":
                if token:
                    headers["Authorization"] = f"Bearer {token}"
                else:
                    return FlowMessage(
                        error_text="Token is required for Token Authentication",
                        step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
                    )

            if http_method not in ("GET", "POST", "PUT"):
                return FlowMessage(
                    error_text="Unsupported HTTP method. Supported methods are GET, POST, and PUT.",
                    step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
                )

            if http_method == "GET":
                response = requests.get(api_url, headers=headers, stream=True)
            elif http_method == "POST":
                response = requests.post(api_url, headers=headers, stream=True)
            elif http_method == "PUT":
                response = requests.put(api_url, headers=headers, stream=True)

            if response.status_code >= 200 and response.status_code < 300:
                payload = None
                if self.configs["response_destination"] in ("flow_message", "both"):
                    payload = response.json()
                    with output_file_path.open("wb") as output_file:
                        output_file.write(response.content)
                else:
                    with output_file_path.open("wb") as output_file:
                        for chunk in response.iter_content(chunk_size=1024):
                            if chunk:
                                output_file.write(chunk)

                return FlowMessage(output_text="Response downloaded and saved successfully.", payload=payload)
            else:
                return FlowMessage(
                    error_text=f"Failed to fetch response. Status Code: {response.status_code}",
                    step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
                )
        except Exception as e:
            return FlowMessage(
                error_text=f"An error occurred: {e!s}",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )



"""Should contain methods to validate Graph Transformation Rules sheet,
as well App Data Model (RDF)
"""

import logging
from typing import Any


def _find_circular_reference_path(
    asset: dict[str, Any], assets: dict[str, dict[str, Any]], max_hierarchy_depth: int = 10000
) -> list:
    original_external_id = asset.get("external_id", "")
    circle: list[str] = [original_external_id]
    parent_external_id = asset.get("parent_external_id")
    if isinstance(parent_external_id, str):
        ref = assets.get(parent_external_id)
    else:
        ref = None

    hop = 0
    while ref is not None and hop < max_hierarchy_depth:
        hop += 1
        if external_id := ref.get("external_id"):
            circle.append(external_id)
        if len(circle) != len(set(circle)):
            msg = (
                f"Found circular reference in asset hierarchy which starts with "
                f"{original_external_id} and enters loop at {circle[-1]}. "
            )
            logging.error(msg)
            return circle
        if parent_external_id := ref.get("parent_external_id"):
            ref = assets.get(parent_external_id)
        else:
            ref = None

    if hop >= max_hierarchy_depth:
        msg = (
            f"Your asset hierarchy is too deep. Max depth is {max_hierarchy_depth}. "
            "You probably have a circular reference."
        )
        logging.error(msg)
        return circle
    else:
        return []


def validate_asset_hierarchy(
    assets: dict[str, dict[str, Any]]
) -> tuple[list[str], list[list[str]], dict[str, list[str]]]:
    """Validates asset hierarchy and reports on orphan assets and circular dependency

    Args:
        assets : A dictionary of assets with external_id as key

    Returns:
        List of orphan assets external ids and list of circular path of external ids.
        If both lists are empty, the hierarchy is healthy.
    """
    orphan_assets: list[str] = []
    circular_reference_paths: list[list[str]] = []
    parent_children_map: dict[str, list[str]] = {}

    for asset in assets.values():
        parent_external_id = asset.get("parent_external_id")
        asset_extarnal_id = asset.get("external_id")
        if asset_extarnal_id and parent_external_id:
            if parent_external_id in parent_children_map:
                parent_children_map[parent_external_id].append(asset_extarnal_id)
            else:
                parent_children_map[parent_external_id] = [asset_extarnal_id]
        if parent_external_id is not None and parent_external_id not in assets:
            msg = (
                f"Found orphan asset {asset.get('external_id')} with parent {parent_external_id} which does not exist."
            )
            logging.error(msg)
            if external_id := asset.get("external_id"):
                orphan_assets.append(external_id)
        circular_reference_path = _find_circular_reference_path(asset, assets)
        if not len(circular_reference_path):
            continue

        # Save the circle only once, not once for every asset
        if set(circular_reference_path) in [set(path) for path in circular_reference_paths]:
            continue
        circular_reference_paths.append(circular_reference_path)
    return orphan_assets, circular_reference_paths, parent_children_map



import logging
import sys
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, fields
from datetime import datetime
from itertools import groupby
from typing import Any, Literal, overload

from cognite.client import CogniteClient
from cognite.client.data_classes import AssetWrite, LabelDefinitionWrite, RelationshipWrite, RelationshipWriteList
from pydantic_core import ErrorDetails
from rdflib.query import ResultRow

from cognite.neat.graph.stores import NeatGraphStoreBase
from cognite.neat.rules.models import Rules
from cognite.neat.rules.models.rules import Property
from cognite.neat.utils import remove_namespace
from cognite.neat.utils.utils import epoch_now_ms

from ._base import CogniteLoader

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc

METADATA_VALUE_MAX_LENGTH = 5120


@dataclass(frozen=True)
class AssetLoaderMetadataKeys:
    """Class holding mapping between NEAT metadata key names and their desired names
    in CDF Asset metadata

    Args:
        start_time: Start time key name
        end_time: End time key name
        update_time: Update time key name
        resurrection_time: Resurrection time key name
        identifier: Identifier key name
        active: Active key name
        type: Type key name
    """

    start_time: str = "start_time"
    end_time: str = "end_time"
    update_time: str = "update_time"
    resurrection_time: str = "resurrection_time"
    identifier: str = "identifier"
    active: str = "active"
    type: str = "type"

    def as_aliases(self) -> dict[str, str]:
        return {str(field.default): getattr(self, field.name) for field in fields(self)}


AssetResource = AssetWrite | RelationshipWrite | LabelDefinitionWrite


class AssetLoader(CogniteLoader[AssetResource]):
    """The Asset Loader is used to load a triple store into the CDF core Asset Hierarchy model.

    Args:
        rules: The transformation rules to use
        graph_store: The graph store to load from
        data_set_id: The data set id to load into
        label_data_set_id: The data set id to load labels into. If not set, the same as data_set_id will be used.
        use_orphanage: Whether to use an orphanage for assets without a parent
        use_labels: Whether to use labels attached to the assets and relationships.
        asset_external_id_prefix: Prefix to add to all external ids
        default_metadata_value: Default metadata value to use for assets without a value. If set to None,
            the metadata key will be omitted. Setting this to an empty string will set the metadata key to an empty
             string, thus ensuring that all assets have the metadata keys.
        metadata_keys: Metadata key names to use
        always_store_in_metadata: Whether to store all properties in metadata. This will be the same as setting
            resource_type_property to metadata for all properties. For example, if you have the property `Terminal.name`
            set with resource_type_property = ["name"], then the property will be stored in metadata as
            `Terminal.name=<Value>`. This also includes properties that are relationships.
    """

    # This is guaranteed ot be in the data
    _identifier: str = "identifier"
    # This label is added to all assets if use_labels is True
    _non_historic_label: str = "non-historic"

    def __init__(
        self,
        rules: Rules,
        graph_store: NeatGraphStoreBase,
        data_set_id: int,
        label_data_set_id: int | None = None,
        use_orphanage: bool = True,
        use_labels: bool = True,
        asset_external_id_prefix: str | None = None,
        default_metadata_value: str | None = "",
        metadata_keys: AssetLoaderMetadataKeys | None = None,
        always_store_in_metadata: bool = False,
    ):
        super().__init__(rules, graph_store)
        self._data_set_id = data_set_id
        self._label_data_set_id = label_data_set_id or data_set_id
        self._use_labels = use_labels
        self._use_orphanage = use_orphanage
        self._orphanage_external_id = (
            f"{asset_external_id_prefix or ''}orphanage-{data_set_id}" if use_orphanage else None
        )
        self._asset_external_id_prefix = asset_external_id_prefix
        self._default_metadata_value = default_metadata_value
        self._metadata_keys = metadata_keys or AssetLoaderMetadataKeys()
        # This is used in a hot loop, so we cache it
        self._metadata_key_aliases = self._metadata_keys.as_aliases()
        self._always_store_in_metadata = always_store_in_metadata

        # State:
        self._loaded_assets: set[str] = set()
        self._loaded_labels: set[str] = set()

    @overload
    def load(self, stop_on_exception: Literal[True]) -> Iterable[AssetResource]:
        ...

    @overload
    def load(self, stop_on_exception: Literal[False] = False) -> Iterable[AssetResource | ErrorDetails]:
        ...

    def load(self, stop_on_exception: bool = False) -> Iterable[AssetResource | ErrorDetails]:
        if self.rules.metadata.namespace is None:
            raise ValueError("Namespace must be provided in transformation rules!")
        namespace = self.rules.metadata.namespace

        properties_by_class_name: dict[str, list[Property]] = defaultdict(list)
        for prop in self.rules.properties.values():
            properties_by_class_name[prop.class_id].append(prop)

        if self._use_labels:
            self._loaded_labels.add(self._non_historic_label)
            yield LabelDefinitionWrite(
                external_id=self._non_historic_label, name=self._non_historic_label, data_set_id=self._label_data_set_id
            )

        # Todo Extend rules to sort topological sorting to ensure that parents are loaded before children
        counter = 0
        for class_name in self.rules.classes.keys():
            if self._use_labels:
                self._loaded_labels.add(class_name)
                yield LabelDefinitionWrite(external_id=class_name, name=class_name, data_set_id=self._label_data_set_id)

            class_uri = namespace[class_name]
            logging.debug(f"Processing class <{class_uri}>.")

            try:
                result = self.graph_store.queries.list_instances_of_type(class_uri)
            except Exception as e:
                logging.error(f"Error while querying for instances of class <{class_uri}> into cache. Reason: {e}")
                if stop_on_exception:
                    raise e
                yield ErrorDetails(
                    input={"class_uri": class_uri},
                    loc=("AssetLoader", "load"),
                    msg=f"Error while querying instances of class <{class_uri}> into cache. Reason: {e}",
                    type=f"Exception of type {type(e).__name__} occurred when processing instance of {class_name}",
                )
                continue

            logging.debug(f"Class <{class_name}> has {len(result)} instances")

            for instance_uri, properties_values in groupby(result, lambda x: x[0]):
                instance_id = remove_namespace(instance_uri)
                values_by_property = self._prepare_instance_data(instance_id, properties_values)
                try:
                    asset = self._load_asset(
                        properties_by_class_name[class_name],
                        values_by_property,
                        instance_id,
                        class_name,
                    )
                except Exception as e:
                    logging.error(f"Error while loading asset from instance <{instance_id}>. Reason: {e}")
                    if stop_on_exception:
                        raise e
                    yield ErrorDetails(
                        input={"instance_id": instance_id},
                        loc=("AssetLoader", "load"),
                        msg=f"Error while loading asset from <{instance_id}>. Reason: {e}",
                        type=f"Exception of type {type(e).__name__} occurred when processing instance of {class_name}",
                    )
                    continue
                else:
                    # We know that external_id is always set
                    self._loaded_assets.add(asset.external_id)  # type: ignore[arg-type]
                    yield asset

                try:
                    relationships = self._load_relationships(
                        properties_by_class_name[class_name],
                        values_by_property,
                        remove_namespace(instance_id),
                        class_name,
                    )
                except Exception as e:
                    logging.error(f"Error while loading relationships from instance <{instance_id}>. Reason: {e}")
                    if stop_on_exception:
                        raise e
                    yield ErrorDetails(
                        input={"instance_id": instance_id},
                        loc=("AssetLoader", "load"),
                        msg=f"Error while loading relationships from <{instance_id}>. Reason: {e}",
                        type=f"Exception of type {type(e).__name__} occurred when processing instance of {class_name}",
                    )
                    continue
                else:
                    for relationship in relationships:
                        for label in relationship.labels or []:
                            if label.external_id and label.external_id not in self._loaded_labels:
                                self._loaded_labels.add(label.external_id)
                                yield LabelDefinitionWrite(
                                    name=label.external_id,
                                    external_id=label.external_id,
                                    data_set_id=self._label_data_set_id,
                                )

                    yield from relationships

                counter += 1
                # log every 10000 assets
                if counter % 10_000 == 0:
                    logging.info(" Next 10,000 Assets processed")

            logging.debug(f"Class <{class_name}> processed")

        # Todo Move orphanage to the beginning to ensure that it is loaded first.
        if self._use_orphanage and self._orphanage_external_id not in self._loaded_assets:
            logging.warning(f"Orphanage with external id {self._orphanage_external_id} not found in asset hierarchy!")
            logging.warning(f"Adding default orphanage with external id {self._orphanage_external_id}")
            now = str(datetime.now(UTC))
            if self._use_labels:
                yield LabelDefinitionWrite(
                    external_id="Orphanage", name="Orphanage", data_set_id=self._label_data_set_id
                )

            yield AssetWrite(
                external_id=self._orphanage_external_id,
                name="Orphanage",
                data_set_id=self._data_set_id,
                parent_external_id=None,
                description="Used to store all assets which parent does not exist",
                labels=["Orphanage", self._non_historic_label] if self._use_labels else None,
                metadata={
                    self._metadata_keys.type: "Orphanage",
                    "cdfResourceType": "Asset",
                    self._metadata_keys.start_time: now,
                    self._metadata_keys.update_time: now,
                    self._metadata_keys.identifier: "orphanage",
                    self._metadata_keys.active: "true",
                },
            )

    @overload
    def load_assets(self, stop_on_exception: Literal[True]) -> Iterable[AssetWrite]:
        ...

    @overload
    def load_assets(self, stop_on_exception: Literal[False] = False) -> Iterable[AssetWrite | ErrorDetails]:
        ...

    def load_assets(self, stop_on_exception: Literal[True, False] = False) -> Iterable[AssetWrite | ErrorDetails]:
        for asset_resource in self.load(stop_on_exception):
            if isinstance(asset_resource, AssetWrite):
                yield asset_resource

    @overload
    def load_relationships(self, stop_on_exception: Literal[True]) -> Iterable[RelationshipWrite]:
        ...

    @overload
    def load_relationships(
        self, stop_on_exception: Literal[False] = False
    ) -> Iterable[RelationshipWrite | ErrorDetails]:
        ...

    def load_relationships(
        self, stop_on_exception: Literal[True, False] = False
    ) -> Iterable[RelationshipWrite | ErrorDetails]:
        for asset_resource in self.load(stop_on_exception):
            if isinstance(asset_resource, RelationshipWrite):
                yield asset_resource

    def load_to_cdf(
        self, client: CogniteClient, batch_size: int | None = 1000, max_retries: int = 1, retry_delay: int = 3
    ) -> None:
        raise NotImplementedError

    def _load_asset(
        self, properties: list[Property], values_by_property: dict[str, str | list[str]], instance_id, class_name: str
    ) -> AssetWrite:
        """Converts a set of properties and values to an AssetWrite object."""
        asset_raw = self._load_asset_data(properties, values_by_property, instance_id, class_name)

        # Loading Asset assumes camel case.
        # Todo Change rules to use camel case?
        for snake_case, came_case in [
            ("external_id", "externalId"),
            ("parent_external_id", "parentExternalId"),
            ("geo_location", "geoLocation"),
        ]:
            if snake_case in asset_raw:
                asset_raw[came_case] = asset_raw.pop(snake_case)

        return AssetWrite.load(asset_raw)

    def _load_asset_data(
        self,
        properties: list[Property],
        values_by_property: dict[str, str | list[str]],
        instance_id: str,
        class_name: str,
    ) -> dict[str, Any]:
        """Creates a raw asset dict from a set of properties and values."""
        asset_raw, missing_metadata_properties = self._load_instance_data(properties, values_by_property)

        self._append_missing_properties(asset_raw, missing_metadata_properties, instance_id, class_name)

        asset_raw["dataSetId"] = self._data_set_id

        # Neat specific metadata keys
        if self._use_labels:
            asset_raw["labels"] = [class_name, self._non_historic_label]
        now = str(datetime.now(UTC))
        asset_raw["metadata"][self._metadata_keys.start_time] = now
        asset_raw["metadata"][self._metadata_keys.update_time] = now
        asset_raw["metadata"][self._metadata_keys.identifier] = instance_id
        asset_raw["metadata"][self._metadata_keys.active] = "true"
        asset_raw["metadata"][self._metadata_keys.type] = class_name

        # Rename metadata keys based on configuration
        asset_raw["metadata"] = {self._metadata_key_aliases.get(k, k): v for k, v in asset_raw["metadata"].items()}
        return asset_raw

    def _prepare_instance_data(
        self, instance_id: str, properties_values: Iterable[ResultRow]
    ) -> dict[str, str | list[str]]:
        """Groups the properties and values by property type.

        Returns:
            A dictionary with property type as key and a list of values as value.
        """
        properties_value_tuples: list[tuple[str, str]] = [
            remove_namespace(prop, value) for _, prop, value in properties_values  # type: ignore[misc]
        ]
        # We add an identifier which will be used as fallback for external_id
        properties_value_tuples.append((self._identifier, remove_namespace(instance_id)))
        values_by_property: dict[str, str | list[str]] = {}
        for prop, values in groupby(sorted(properties_value_tuples), lambda x: x[0]):
            values_list: list[str] = [value for _, value in values]  # type: ignore[misc, has-type]
            if len(values_list) == 1:
                values_by_property[prop] = values_list[0]
            else:
                values_by_property[prop] = values_list
        return values_by_property

    def _load_instance_data(
        self, properties: list[Property], values_by_property: dict[str, str | list[str]]
    ) -> tuple[dict[str, Any], set[str]]:
        """This function loads the instance data into a raw asset dict. It also returns a set of metadata keys that
        were not found in the instance data.

        Returns:
            A tuple with the raw asset dict and a set of metadata keys that were not found in the instance data.
        """
        asset_raw: dict[str, Any] = {"metadata": {}}
        missing_metadata_properties: set[str] = set()
        for prop in properties:
            if prop.property_name is None:
                continue
            is_relationship = "Asset" not in prop.cdf_resource_type
            if not self._always_store_in_metadata and is_relationship:
                continue

            if prop.property_name not in values_by_property:
                for property_type in prop.resource_type_property or []:
                    if property_type.casefold() == "metadata":
                        missing_metadata_properties.add(prop.property_name)
                    elif self._always_store_in_metadata:
                        missing_metadata_properties.add(prop.property_name)
                continue
            values = values_by_property[prop.property_name]
            for property_type in prop.resource_type_property or []:
                if property_type.casefold() == "metadata":
                    asset_raw["metadata"][prop.property_name] = self._to_metadata_value(values)
                else:
                    if property_type not in asset_raw:
                        asset_raw[property_type] = values
                    if self._always_store_in_metadata and prop.property_name not in asset_raw["metadata"]:
                        asset_raw["metadata"][prop.property_name] = self._to_metadata_value(values)
            if is_relationship:
                asset_raw["metadata"][prop.property_name] = self._to_metadata_value(values)

        return asset_raw, missing_metadata_properties

    def _append_missing_properties(
        self, asset_raw: dict[str, Any], missing_metadata_properties: set[str], identifier: str, class_name: str
    ) -> None:
        """This function ensures that the raw asset dict has all the required properties such as external_id, name,
        and parent_external_id. It also ensures that the metadata dict has all the required keys."""

        if "external_id" not in asset_raw:
            msg = f"Missing external_id for {class_name} instance {identifier}. Using value <{identifier}>."
            logging.debug(msg)
            asset_raw["external_id"] = identifier
        elif "external_id" in asset_raw and isinstance(asset_raw["external_id"], list):
            external_ids = asset_raw["external_id"]
            msg = (
                f"Multiple values for {class_name} instance {identifier} external_id. "
                f"Using the first one <{external_ids[0]}>."
            )
            logging.debug(msg)
            asset_raw["external_id"] = external_ids[0]

        if self._asset_external_id_prefix:
            asset_raw["external_id"] = f"{self._asset_external_id_prefix}{asset_raw['external_id']}"

        external_id = asset_raw["external_id"]

        if "name" not in asset_raw:
            msg = f"Missing name for {class_name} instance {external_id}. Using value <{identifier}>."
            logging.debug(msg)
            asset_raw["name"] = identifier
        elif "name" in asset_raw and isinstance(asset_raw["name"], list):
            msg = f"Multiple values for {class_name} instance {external_id} name. Joining them into one."
            logging.debug(msg)
            asset_raw["name"] = self._to_metadata_value(asset_raw["name"])

        if "description" in asset_raw and isinstance(asset_raw["description"], list):
            msg = f"Multiple values for {class_name} instance {external_id} description. Joining them into one."
            logging.debug(msg)
            asset_raw["description"] = self._to_metadata_value(asset_raw["description"])

        if "parent_external_id" not in asset_raw and self._orphanage_external_id is not None:
            asset_raw["parent_external_id"] = self._orphanage_external_id
        elif "parent_external_id" in asset_raw and isinstance(asset_raw["parent_external_id"], list):
            msg = (
                f"Multiple values for {class_name} instance {external_id} parent_external_id. "
                f"Using the first one <{asset_raw['parent_external_id'][0]}>."
            )
            logging.debug(msg)
            asset_raw["parent_external_id"] = asset_raw["parent_external_id"][0]

        if "parent_external_id" in asset_raw:
            asset_raw["parent_external_id"] = f"{self._asset_external_id_prefix or ''}{asset_raw['parent_external_id']}"

        for metadata_key in missing_metadata_properties:
            msg = f"{external_id} of type {class_name} is missing metadata key {metadata_key}."
            logging.debug(msg)
            if self._default_metadata_value is not None and metadata_key not in asset_raw["metadata"]:
                asset_raw["metadata"][metadata_key] = self._default_metadata_value
                logging.debug(f"\tKey {metadata_key} added to <{external_id}> metadata!")

    @staticmethod
    def _to_metadata_value(values: str | list[str]) -> str:
        """A helper function to convert a list of values to a metadata value string respecting metadata value length."""
        # Sorting for deterministic results
        return values if isinstance(values, str) else ", ".join(sorted(values))[: METADATA_VALUE_MAX_LENGTH - 1]

    def _load_relationships(
        self,
        properties: list[Property],
        values_by_property: dict[str, str | list[str]],
        instance_id: str,
        class_name: str,
    ) -> RelationshipWriteList:
        """Converts a set of properties and values to a RelationshipWriteList object."""
        relationships = RelationshipWriteList([])
        relationship_properties = [prop for prop in properties if "Relationship" in prop.cdf_resource_type]
        epoch_now = epoch_now_ms()
        for prop in relationship_properties:
            if not prop.property_name:
                continue
            values = values_by_property.get(prop.property_name, [])
            if not values:
                continue
            value_list = [values] if isinstance(values, str) else values
            for value in value_list:
                relationship = self._load_relationship(prop, value, instance_id, class_name, epoch_now)
                relationships.append(relationship)

        return relationships

    def _load_relationship(
        self,
        prop: Property,
        target_id: str,
        instance_id: str,
        class_name: str,
        epoch_now: int,
    ) -> RelationshipWrite:
        """Converts a set of properties and values to a RelationshipWrite object."""
        prefix = self._asset_external_id_prefix or ""
        labels = (
            [class_name, self._non_historic_label, prop.expected_value_type.suffix, prop.property_id]
            if self._use_labels
            else None
        )
        relationship_raw = {
            "externalId": f"{prefix}{instance_id}:{prefix}{target_id}",
            "sourceExternalId": f"{prefix}{instance_id}",
            "targetExternalId": f"{prefix}{target_id}",
            "sourceType": prop.source_type,
            "targetType": prop.target_type,
            "dataSetId": self._data_set_id,
            "labels": labels,
            "startTime": epoch_now,
        }
        return RelationshipWrite.load(relationship_raw)



import re
from collections.abc import Callable
from typing import Annotated, Any, cast

import rdflib
from pydantic import (
    AfterValidator,
    BeforeValidator,
    Field,
    HttpUrl,
    StringConstraints,
    TypeAdapter,
    ValidationInfo,
    WrapValidator,
)
from pydantic.functional_serializers import PlainSerializer
from pydantic_core import PydanticCustomError

from cognite.neat.rules import exceptions

from ._base import (
    CLASS_ID_COMPLIANCE_REGEX,
    ENTITY_ID_REGEX_COMPILED,
    MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX,
    PREFIX_COMPLIANCE_REGEX,
    PROPERTY_ID_COMPLIANCE_REGEX,
    VERSION_COMPLIANCE_REGEX,
    VERSIONED_ENTITY_REGEX_COMPILED,
    ClassEntity,
    ContainerEntity,
    ParentClassEntity,
    ReferenceEntity,
    Undefined,
    Unknown,
    ViewEntity,
    ViewPropEntity,
)
from ._value import DMS_VALUE_TYPE_MAPPINGS, XSD_VALUE_TYPE_MAPPINGS, DMSValueType, XSDValueType


def _custom_error(exc_factory: Callable[[str | None, Exception], Any]) -> Any:
    def _validator(value: Any, next_: Any, ctx: ValidationInfo) -> Any:
        try:
            return next_(value, ctx)
        except Exception:
            raise exc_factory(ctx.field_name, value) from None

    return WrapValidator(_validator)


def _raise(exception: PydanticCustomError):
    raise exception


def _split_parent(value: str | list[ParentClassEntity]) -> list[ParentClassEntity] | None:
    if isinstance(value, list) and all(isinstance(x, ParentClassEntity) for x in value):
        return value

    if not (isinstance(value, str) and value):
        return None

    parents = []
    for v in value.replace(", ", ",").split(","):
        if ENTITY_ID_REGEX_COMPILED.match(v) or VERSIONED_ENTITY_REGEX_COMPILED.match(v):
            parents.append(ParentClassEntity.from_string(entity_string=v))
        else:
            parents.append(ParentClassEntity(prefix=Undefined, suffix=v, name=v))

    return parents


def _check_parent(value: list[ParentClassEntity]) -> list[ParentClassEntity]:
    if not value:
        return value
    if illegal_ids := [v for v in value if re.search(MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX, v.suffix)]:
        raise exceptions.MoreThanOneNonAlphanumericCharacter(
            "parent", ", ".join(cast(list[str], illegal_ids))
        ).to_pydantic_custom_error()
    if illegal_ids := [v for v in value if not re.match(CLASS_ID_COMPLIANCE_REGEX, v.suffix)]:
        for v in illegal_ids:
            print(v.id)
        raise exceptions.ClassSheetParentClassIDRegexViolation(
            cast(list[str], illegal_ids), CLASS_ID_COMPLIANCE_REGEX
        ).to_pydantic_custom_error()
    return value


StrOrListType = Annotated[
    str | list[str],
    BeforeValidator(lambda value: value.replace(", ", ",").split(",") if isinstance(value, str) and value else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]) if isinstance(value, list) else value,
        return_type=str,
        when_used="unless-none",
    ),
]


StrListType = Annotated[
    list[str],
    BeforeValidator(lambda value: [entry.strip() for entry in value.split(",")] if isinstance(value, str) else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]), return_type=str, when_used="unless-none"
    ),
]

NamespaceType = Annotated[
    rdflib.Namespace,
    BeforeValidator(
        lambda value: (
            rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(value))
            if value.endswith("#") or value.endswith("/")
            else rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(f"{value}/"))
        )
    ),
]

PrefixType = Annotated[
    str,
    StringConstraints(pattern=PREFIX_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: exceptions.PrefixesRegexViolation(
            cast(list[str], [value]), PREFIX_COMPLIANCE_REGEX
        ).to_pydantic_custom_error()
    ),
]

ExternalIdType = Annotated[
    str,
    Field(min_length=1, max_length=255),
]

VersionType = Annotated[
    str,
    StringConstraints(pattern=VERSION_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: exceptions.VersionRegexViolation(
            version=cast(str, value), regex_expression=VERSION_COMPLIANCE_REGEX
        ).to_pydantic_custom_error()
    ),
]


ParentClassType = Annotated[
    list[ParentClassEntity] | None,
    BeforeValidator(_split_parent),
    AfterValidator(_check_parent),
    PlainSerializer(
        lambda v: ",".join([entry.versioned_id for entry in v]) if v else None,
        return_type=str,
        when_used="unless-none",
    ),
]

ClassType = Annotated[
    ClassEntity,
    BeforeValidator(lambda value: (ClassType.from_raw(value) if isinstance(value, str) else value)),
    AfterValidator(
        lambda value: (
            _raise(exceptions.MoreThanOneNonAlphanumericCharacter("class_", value.suffix).to_pydantic_custom_error())
            if re.search(MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX, value.suffix)
            else (
                value
                if re.match(CLASS_ID_COMPLIANCE_REGEX, value.suffix)
                else _raise(
                    exceptions.ClassSheetClassIDRegexViolation(
                        value.suffix, CLASS_ID_COMPLIANCE_REGEX
                    ).to_pydantic_custom_error()
                )
            )
        )
    ),
    PlainSerializer(
        lambda v: v.versioned_id,
        return_type=str,
        when_used="unless-none",
    ),
]


PropertyType = Annotated[
    str,
    AfterValidator(
        lambda value: (
            _raise(exceptions.MoreThanOneNonAlphanumericCharacter("property", value).to_pydantic_custom_error())
            if re.search(MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX, value)
            else (
                value
                if re.match(PROPERTY_ID_COMPLIANCE_REGEX, value)
                else _raise(
                    exceptions.PropertyIDRegexViolation(value, PROPERTY_ID_COMPLIANCE_REGEX).to_pydantic_custom_error()
                )
            )
        )
    ),
]


def _reference_entity_type_before_validator(value: Any | None = None) -> Any:
    if not value:
        return None
    elif isinstance(value, rdflib.URIRef | ReferenceEntity):
        return value
    elif ReferenceEntity.from_raw(value).prefix == Undefined:
        # case1: then this must be a URIRef!
        return rdflib.URIRef(str(TypeAdapter(HttpUrl).validate_python(value)))
    else:
        # case2: this is a ReferenceEntity
        return ReferenceEntity.from_raw(value)


ReferenceType = Annotated[
    ReferenceEntity | rdflib.URIRef | None,
    BeforeValidator(_reference_entity_type_before_validator),
    PlainSerializer(
        lambda v: v.versioned_id if isinstance(v, ReferenceEntity) else str(v),
        return_type=str,
        when_used="unless-none",
    ),
]


ContainerType = Annotated[
    ContainerEntity,
    BeforeValidator(ContainerEntity.from_raw),
    PlainSerializer(
        lambda v: v.versioned_id,
        return_type=str,
        when_used="unless-none",
    ),
]


ViewType = Annotated[
    ViewEntity,
    BeforeValidator(ViewEntity.from_raw),
    PlainSerializer(
        lambda v: v.versioned_id,
        return_type=str,
        when_used="unless-none",
    ),
]


def _semantic_value_type_before_validator(value: Any) -> Any:
    if isinstance(value, XSDValueType | ClassEntity) or not isinstance(value, str):
        return value
    elif value in XSD_VALUE_TYPE_MAPPINGS:
        return XSD_VALUE_TYPE_MAPPINGS[value]
    elif value.lower() in XSD_VALUE_TYPE_MAPPINGS:
        return XSD_VALUE_TYPE_MAPPINGS[value.lower()]
    elif value == "#N/A":
        return ClassEntity(prefix=Undefined, suffix=Unknown)
    else:
        return ClassEntity.from_raw(value)


def _semantic_value_type_serializer(value: Any) -> str:
    if isinstance(value, ClassEntity):
        return value.versioned_id
    elif isinstance(value, XSDValueType):
        return value.suffix
    else:
        return str(value)


def _dms_value_type_before_validator(value: Any) -> Any:
    if isinstance(value, DMSValueType | ViewPropEntity) or not isinstance(value, str):
        return value
    elif value in DMS_VALUE_TYPE_MAPPINGS:
        return DMS_VALUE_TYPE_MAPPINGS[value]
    elif value.lower() in DMS_VALUE_TYPE_MAPPINGS:
        return DMS_VALUE_TYPE_MAPPINGS[value.lower()]
    else:
        return ViewPropEntity.from_raw(value)


SemanticValueType = Annotated[
    XSDValueType | ClassEntity,
    BeforeValidator(_semantic_value_type_before_validator),
    PlainSerializer(
        _semantic_value_type_serializer,
        return_type=str,
        when_used="unless-none",
    ),
]

CdfValueType = Annotated[
    DMSValueType | ViewPropEntity,
    BeforeValidator(_dms_value_type_before_validator),
    PlainSerializer(
        lambda v: v.versioned_id if isinstance(v, ViewPropEntity | ViewEntity) else v.dms()._type,
        return_type=str,
        when_used="unless-none",
    ),
]


def _from_str_or_list_container(value: Any) -> list[ContainerEntity] | Any:
    if not value:
        return value
    if isinstance(value, str):
        return [ContainerEntity.from_raw(entry.strip()) for entry in value.split(",")]
    elif isinstance(value, list):
        return [ContainerEntity.from_raw(entry.strip()) if isinstance(entry, str) else entry for entry in value]
    else:
        return value


def _from_str_or_list_view(value: Any) -> list[ViewEntity] | Any:
    if not value:
        return value
    if isinstance(value, str):
        return [ViewEntity.from_raw(entry.strip()) for entry in value.split(",")]
    elif isinstance(value, list):
        return [ViewEntity.from_raw(entry.strip()) if isinstance(entry, str) else entry for entry in value]
    else:
        return value


ContainerListType = Annotated[
    list[ContainerEntity],
    BeforeValidator(_from_str_or_list_container),
    PlainSerializer(
        lambda v: ",".join([entry.versioned_id for entry in v]),
        return_type=str,
        when_used="unless-none",
    ),
]

ViewListType = Annotated[
    list[ViewEntity],
    BeforeValidator(_from_str_or_list_view),
    PlainSerializer(
        lambda v: ",".join([entry.versioned_id for entry in v]),
        return_type=str,
        when_used="unless-none",
    ),
]



import json
import zipfile
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Literal, overload

from pydantic import ValidationError

from cognite.neat.rules import issues
from cognite.neat.rules._shared import Rules
from cognite.neat.rules.importers._base import BaseImporter, _handle_issues
from cognite.neat.rules.importers._dtdl2rules.dtdl_converter import _DTDLConverter
from cognite.neat.rules.importers._dtdl2rules.spec import DTDL_CLS_BY_TYPE_BY_SPEC, DTDLBase, Interface
from cognite.neat.rules.issues import IssueList, ValidationIssue
from cognite.neat.rules.models._rules import InformationRules, RoleTypes
from cognite.neat.rules.models._rules.base import SchemaCompleteness, SheetList
from cognite.neat.rules.models._rules.information_rules import InformationClass, InformationProperty
from cognite.neat.utils.text import to_pascal


class DTDLImporter(BaseImporter):
    """Importer for DTDL (Digital Twin Definition Language).

    This importer supports DTDL v2.0 and v3.0.

    It is recommended to use the class methods `from_directory` and `from_zip` to create an instance of this class.

    Args:
        items (Sequence[DTDLBase]): A sequence of DTDLBase objects.
        title (str, optional): Title of the data model. Defaults to None.
        read_issues (list[ValidationIssue], optional): A list of issues that occurred during reading. Defaults to None.
        schema (SchemaCompleteness, optional): Schema completeness. Defaults to SchemaCompleteness.partial.

    """

    def __init__(
        self,
        items: Sequence[DTDLBase],
        title: str | None = None,
        read_issues: list[ValidationIssue] | None = None,
        schema: SchemaCompleteness = SchemaCompleteness.partial,
    ) -> None:
        self._items = items
        self.title = title
        self._read_issues = read_issues
        self._schema_completeness = schema

    @classmethod
    def _from_file_content(cls, file_content: str, filepath: Path) -> Iterable[DTDLBase | ValidationIssue]:
        raw = json.loads(file_content)
        if isinstance(raw, dict):
            if (context := raw.get("@context")) is None:
                yield issues.fileread.InvalidFileFormatWarning(filepath=filepath, reason="Missing '@context' key.")
                return
            raw_list = [raw]
        elif isinstance(raw, list):
            context = next(
                (entry["@context"] for entry in raw if isinstance(entry, dict) and "@context" in entry), None
            )
            if context is None:
                yield issues.fileread.InvalidFileFormatWarning(filepath=filepath, reason="Missing '@context' key.")
                return
            raw_list = raw
        else:
            yield issues.fileread.InvalidFileFormatWarning(
                filepath=filepath, reason="Content is not an object or array."
            )
            return

        if isinstance(context, list):
            context = context[0]
        Interface.default_context = context
        spec_version = context.split(";")[1]
        try:
            cls_by_type = DTDL_CLS_BY_TYPE_BY_SPEC[spec_version]
        except KeyError:
            yield issues.fileread.UnsupportedSpecWarning(filepath=filepath, version=spec_version, spec_name="DTDL")
            return

        for item in raw_list:
            if not (type_ := item.get("@type")):
                yield issues.fileread.InvalidFileFormatWarning(filepath=filepath, reason="Missing '@type' key.")
                continue
            cls_ = cls_by_type.get(type_)
            if cls_ is None:
                yield issues.fileread.UnknownItemWarning(reason=f"Unknown '@type' {type_}.", filepath=filepath)
                continue
            try:
                yield cls_.model_validate(item)
            except ValidationError as e:
                yield issues.fileread.InvalidFileFormatWarning(filepath=filepath, reason=str(e))
            except Exception as e:
                yield issues.fileread.BugInImporterWarning(filepath=filepath, error=str(e), importer_name=cls.__name__)

    @classmethod
    def from_directory(cls, directory: Path) -> "DTDLImporter":
        items: list[DTDLBase] = []
        issues: list[ValidationIssue] = []
        for filepath in directory.glob("**/*.json"):
            for item in cls._from_file_content(filepath.read_text(), filepath):
                if isinstance(item, ValidationIssue):
                    issues.append(item)
                else:
                    items.append(item)
        return cls(items, directory.stem, read_issues=issues)

    @classmethod
    def from_zip(cls, zip_file: Path) -> "DTDLImporter":
        items: list[DTDLBase] = []
        issues: list[ValidationIssue] = []
        with zipfile.ZipFile(zip_file) as z:
            for filepath in z.namelist():
                if filepath.endswith(".json"):
                    for item in cls._from_file_content(z.read(filepath).decode(), Path(filepath)):
                        if isinstance(item, ValidationIssue):
                            issues.append(item)
                        else:
                            items.append(item)
        return cls(items, zip_file.stem, read_issues=issues)

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None, is_reference: bool = False) -> Rules:
        ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None, is_reference: bool = False
    ) -> tuple[Rules | None, IssueList]:
        ...

    def to_rules(
        self,
        errors: Literal["raise", "continue"] = "continue",
        role: RoleTypes | None = None,
        is_reference: bool = False,
    ) -> tuple[Rules | None, IssueList] | Rules:
        converter = _DTDLConverter(self._read_issues)

        converter.convert(self._items)

        metadata = self._default_metadata()
        metadata["schema"] = self._schema_completeness.value
        if self.title:
            metadata["title"] = to_pascal(self.title)
        try:
            most_common_prefix = converter.get_most_common_prefix()
        except ValueError:
            # No prefixes are defined so we just use the default prefix...
            ...
        else:
            metadata["prefix"] = most_common_prefix
        with _handle_issues(converter.issues) as future:
            rules = InformationRules(
                metadata=metadata,
                properties=SheetList[InformationProperty](data=converter.properties),
                classes=SheetList[InformationClass](data=converter.classes),
            )
        if future.result == "failure":
            if errors == "continue":
                return None, converter.issues
            else:
                raise converter.issues.as_errors()

        return self._to_output(rules, converter.issues, errors, role, is_reference)



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Generic, Literal, TypeVar, overload

from cognite.client import CogniteClient
from pydantic_core import ErrorDetails

from cognite.neat.graph.models import Triple
from cognite.neat.graph.stores import NeatGraphStoreBase
from cognite.neat.graph.transformation.query_generator.sparql import build_construct_query
from cognite.neat.rules.models import Rules

T_Output = TypeVar("T_Output")


class BaseLoader(ABC, Generic[T_Output]):
    """Base class for all loaders.

    A loader is a class that loads data from a source graph into
    target outside Neat.
    """

    def __init__(self, rules: Rules, graph_store: NeatGraphStoreBase):
        self.rules = rules
        self.graph_store = graph_store

    @overload
    def load(self, stop_on_exception: Literal[True]) -> Iterable[T_Output]:
        ...

    @overload
    def load(self, stop_on_exception: Literal[False] = False) -> Iterable[T_Output | ErrorDetails]:
        ...

    @abstractmethod
    def load(self, stop_on_exception: bool = False) -> Iterable[T_Output | ErrorDetails]:
        """Load the graph with data."""
        pass

    def _iterate_class_triples(self, exclude_classes: set[str] | None = None) -> Iterable[tuple[str, Iterable[Triple]]]:
        """Iterate over all classes and their triples."""
        for class_name in self.rules.classes:
            if exclude_classes is not None and class_name in exclude_classes:
                continue
            try:
                sparql_construct_query = build_construct_query(
                    self.graph_store.graph, class_name, self.rules, properties_optional=True
                )
            except Exception as e:
                logging.error(f"Failed to build construct query for class {class_name}: {e}")
                continue

            yield class_name, self.graph_store.query_delayed(sparql_construct_query)


class CogniteLoader(BaseLoader[T_Output], ABC):
    """Base class for all loaders.

    A loader is a class that loads data from a source graph into
    target outside Neat.
    """

    @abstractmethod
    def load_to_cdf(
        self, client: CogniteClient, batch_size: int | None = 1000, max_retries: int = 1, retry_delay: int = 3
    ) -> None:
        """Load the graph with data."""
        pass



from pydantic_core import ErrorDetails, PydanticCustomError


class NeatError(Exception):
    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str

    def to_pydantic_custom_error(self):
        return PydanticCustomError(
            self.type_,
            self.message,
            dict(type_=self.type_, code=self.code, description=self.description, example=self.example, fix=self.fix),
        )

    def to_error_dict(self) -> ErrorDetails:
        return {
            "type": self.type_,
            "loc": (),
            "msg": self.message,
            "input": None,
            "ctx": dict(
                type_=self.type_,
                code=self.code,
                description=self.description,
                example=self.example,
                fix=self.fix,
            ),
        }


class NeatWarning(UserWarning):
    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str



import warnings
from collections.abc import Collection, Iterable
from pathlib import Path
from typing import Literal, TypeAlias, cast

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResourceList
from cognite.client.exceptions import CogniteAPIError

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models._rules import InformationRules
from cognite.neat.rules.models._rules.base import ExtensionCategory, SheetList
from cognite.neat.rules.models._rules.dms_architect_rules import DMSContainer, DMSRules
from cognite.neat.rules.models._rules.dms_schema import DMSSchema, PipelineSchema
from cognite.neat.utils.cdf_loaders import (
    ContainerLoader,
    DataModelingLoader,
    DataModelLoader,
    RawDatabaseLoader,
    RawTableLoader,
    ResourceLoader,
    SpaceLoader,
    TransformationLoader,
    ViewLoader,
)

from ._base import CDFExporter
from ._models import UploadResult

Component: TypeAlias = Literal["all", "spaces", "data_models", "views", "containers"]


class DMSExporter(CDFExporter[DMSSchema]):
    """Class for exporting rules object to CDF Data Model Storage (DMS).

    Args:
        export_components (frozenset[Literal["all", "spaces", "data_models", "views", "containers"]], optional):
            Which components to export. Defaults to frozenset({"all"}).
        include_space (set[str], optional):
            If set, only export components in the given spaces. Defaults to None which means all spaces.
        existing_handling (Literal["fail", "skip", "update", "force"], optional): How to handle existing components.
            Defaults to "update". See below for details.
        export_pipeline (bool, optional): Whether to export the pipeline. Defaults to False. This means setting
            up transformations, RAW databases and tables to populate the data model.
        instance_space (str, optional): The space to use for the instance. Defaults to None.
    ... note::

        - "fail": If any component already exists, the export will fail.
        - "skip": If any component already exists, it will be skipped.
        - "update": If any component already exists, it will be updated.
        - "force": If any component already exists, it will be deleted and recreated.

    """

    def __init__(
        self,
        export_components: Component | Collection[Component] = "all",
        include_space: set[str] | None = None,
        existing_handling: Literal["fail", "skip", "update", "force"] = "update",
        export_pipeline: bool = False,
        instance_space: str | None = None,
    ):
        self.export_components = {export_components} if isinstance(export_components, str) else set(export_components)
        self.include_space = include_space
        self.existing_handling = existing_handling
        self.export_pipeline = export_pipeline
        self.instance_space = instance_space
        self._schema: DMSSchema | None = None

    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        """Export the rules to a file(s).

        If the file is a directory, the components will be exported to separate files, otherwise they will be
        exported to a zip file.

        Args:
            filepath: Directory or zip file path to export to.
            rules:
        """
        if filepath.is_dir():
            self._export_to_directory(filepath, rules)
        else:
            self._export_to_zip_file(filepath, rules)

    def _export_to_directory(self, directory: Path, rules: Rules) -> None:
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_directory(directory, exclude=exclude, new_line=self._new_line, encoding=self._encoding)

    def _export_to_zip_file(self, filepath: Path, rules: Rules) -> None:
        if filepath.suffix not in {".zip"}:
            warnings.warn("File extension is not .zip, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".zip")
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_zip(filepath, exclude=exclude)

    def _create_exclude_set(self):
        if "all" in self.export_components:
            exclude = set()
        else:
            exclude = {"spaces", "data_models", "views", "containers"} - self.export_components
        return exclude

    def export(self, rules: Rules) -> DMSSchema:
        if isinstance(rules, DMSRules):
            dms_rules = rules
        elif isinstance(rules, InformationRules):
            dms_rules = rules.as_dms_architect_rules()
        else:
            raise ValueError(f"{type(rules).__name__} cannot be exported to DMS")

        is_solution_model = (
            dms_rules.reference and dms_rules.metadata.external_id != dms_rules.reference.metadata.external_id
        )
        is_new_model = dms_rules.reference is None
        if is_new_model or is_solution_model:
            return dms_rules.as_schema(self.export_pipeline, self.instance_space)

        # This is an extension of an existing model.
        reference_rules = cast(DMSRules, dms_rules.reference).copy(deep=True)
        reference_schema = reference_rules.as_schema(self.export_pipeline)

        # Todo Move this to an appropriate location
        # Merging Reference with User Rules
        combined_rules = dms_rules.copy(deep=True)
        existing_containers = {container.class_ for container in combined_rules.containers or []}
        if combined_rules.containers is None:
            combined_rules.containers = SheetList[DMSContainer](data=[])
        for container in reference_rules.containers or []:
            if container.class_ not in existing_containers:
                container.reference = None
                combined_rules.containers.append(container)
        existing_views = {view.class_ for view in combined_rules.views}
        for view in reference_rules.views:
            if view.class_ not in existing_views:
                view.reference = None
                combined_rules.views.append(view)
        existing_properties = {(property_.class_, property_.property_) for property_ in combined_rules.properties}
        for property_ in reference_rules.properties:
            if (property_.class_, property_.property_) not in existing_properties:
                property_.reference = None
                combined_rules.properties.append(property_)

        schema = combined_rules.as_schema(self.export_pipeline, self.instance_space)

        if dms_rules.metadata.extension in (ExtensionCategory.addition, ExtensionCategory.reshape):
            # We do not freeze views as they might be changed, even for addition,
            # in case, for example, new properties are added. The validation will catch this.
            schema.frozen_ids.update(set(reference_schema.containers.as_ids()))
            schema.frozen_ids.update(set(reference_schema.node_types.as_ids()))
        return schema

    def export_to_cdf(self, rules: Rules, client: CogniteClient, dry_run: bool = False) -> Iterable[UploadResult]:
        schema = self.export(rules)
        to_export: list[tuple[CogniteResourceList, ResourceLoader]] = []
        if self.export_components.intersection({"all", "spaces"}):
            to_export.append((schema.spaces, SpaceLoader(client)))
        if self.export_components.intersection({"all", "containers"}):
            to_export.append((schema.containers, ContainerLoader(client)))
        if self.export_components.intersection({"all", "views"}):
            to_export.append((schema.views, ViewLoader(client, self.existing_handling)))
        if self.export_components.intersection({"all", "data_models"}):
            to_export.append((schema.data_models, DataModelLoader(client)))
        if isinstance(schema, PipelineSchema):
            to_export.append((schema.databases, RawDatabaseLoader(client)))
            to_export.append((schema.raw_tables, RawTableLoader(client)))
            to_export.append((schema.transformations, TransformationLoader(client)))

        # The conversion from DMS to GraphQL does not seem to be triggered even if the views
        # are changed. This is a workaround to force the conversion.
        redeploy_data_model = False

        for all_items, loader in to_export:
            all_item_ids = loader.get_ids(all_items)
            skipped = sum(1 for item_id in all_item_ids if item_id in schema.frozen_ids)
            item_ids = [item_id for item_id in all_item_ids if item_id not in schema.frozen_ids]
            cdf_items = loader.retrieve(item_ids)
            cdf_item_by_id = {loader.get_id(item): item for item in cdf_items}
            items = [item for item in all_items if loader.get_id(item) in item_ids]
            to_create, to_update, unchanged, to_delete = [], [], [], []
            is_redeploying = loader.resource_name == "data_models" and redeploy_data_model
            for item in items:
                if (
                    isinstance(loader, DataModelingLoader)
                    and self.include_space is not None
                    and not loader.in_space(item, self.include_space)
                ):
                    continue

                cdf_item = cdf_item_by_id.get(loader.get_id(item))
                if cdf_item is None:
                    to_create.append(item)
                elif is_redeploying:
                    to_update.append(item)
                    to_delete.append(cdf_item)
                elif loader.are_equal(item, cdf_item):
                    unchanged.append(item)
                else:
                    to_update.append(item)

            created = len(to_create)
            failed_created = 0

            if self.existing_handling in ["update", "force"]:
                changed = len(to_update)
                failed_changed = 0
            elif self.existing_handling == "skip":
                changed = 0
                failed_changed = 0
                skipped += len(to_update)
            elif self.existing_handling == "fail":
                failed_changed = len(to_update)
                changed = 0
            else:
                raise ValueError(f"Unsupported existing_handling {self.existing_handling}")

            error_messages: list[str] = []
            if not dry_run:
                if to_delete:
                    try:
                        loader.delete(to_delete)
                    except CogniteAPIError as e:
                        error_messages.append(f"Failed delete: {e.message}")

                if isinstance(loader, DataModelingLoader):
                    to_create = loader.sort_by_dependencies(to_create)

                try:
                    loader.create(to_create)
                except CogniteAPIError as e:
                    failed_created = len(e.failed) + len(e.unknown)
                    created -= failed_created
                    error_messages.append(e.message)

                if self.existing_handling in ["update", "force"]:
                    try:
                        loader.update(to_update)
                    except CogniteAPIError as e:
                        failed_changed = len(e.failed) + len(e.unknown)
                        changed -= failed_changed
                        error_messages.append(e.message)

            yield UploadResult(
                name=loader.resource_name,
                created=created,
                changed=changed,
                unchanged=len(unchanged),
                skipped=skipped,
                failed_created=failed_created,
                failed_changed=failed_changed,
                error_messages=error_messages,
            )

            if loader.resource_name == "views" and (created or changed) and not redeploy_data_model:
                redeploy_data_model = True



import logging

import requests
from rdflib import Graph, Namespace
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore

from cognite.neat.constants import DEFAULT_NAMESPACE, PREFIXES

from ._base import NeatGraphStoreBase


class GraphDBStore(NeatGraphStoreBase):
    """GraphDB is a class that stores the graph in a GraphDB instances and provides methods to
    read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        base_prefix : Used as a base prefix for graph namespace, allowing querying graph data using a shortform of a URI
        namespace : Namespace (aka URI) used to resolve any relative URI in the graph
        prefixes : Dictionary of additional prefixes used and bounded to the graph
    """

    rdf_store_type = "graphdb"

    def __init__(
        self,
        graph: Graph | None = None,
        base_prefix: str = "",  # usually empty
        namespace: Namespace = DEFAULT_NAMESPACE,
        prefixes: dict = PREFIXES,
    ):
        super().__init__(graph, base_prefix, namespace, prefixes)
        self.graph_db_rest_url: str = "http://localhost:7200"

    def _set_graph(self) -> None:
        logging.info("Initializing graph store with GraphDB")
        store = SPARQLUpdateStore(
            query_endpoint=self.rdf_store_query_url,
            update_endpoint=self.rdf_store_update_url,
            returnFormat=self.returnFormat,
            context_aware=False,
            postAsEncoded=False,
            autocommit=False,
        )
        self.graph = Graph(store=store)

    def drop(self):
        """Drops the graph."""
        r = requests.delete(f"{self.rdf_store_query_url}/rdf-graphs/service?default")
        logging.info(f"Dropped graph with state: {r.text}")



import logging
from typing import cast

from rdflib import Literal, URIRef

from cognite.neat.graph.stores import NeatGraphStoreBase


def simple_entity_matcher(
    graph_store: NeatGraphStoreBase,
    source_class: str,
    source_property: str,
    source_value_type: str = "single_value_str",
    target_class: str | None = None,
    target_property: str | None = None,
    relationship_name: str = "link",
    link_direction: str = "target_to_source",  # source_to_target, bidirectional
    matching_method: str = "regexp",  # exact_match, similarity
    link_namespace: str = "http://purl.org/cognite/neat#",
) -> int:
    """simple_entity_matcher performs a simple entity matching between two classes in the graph using
    either exact match or regular expression matching.
    The matching is performed on the values of the source_property and target_property.
    The matching is performed in the direction specified by link_direction.
    The matching_method can be either exact_match or regexp.
    If the source_value_type is multi_value_str, the values are split on comma and added as separate triples.
    Args:
        graph_store (NeatGraphStoreBase): The graph store to perform the matching on and add the links to
        source_class (str): The class of the source entities
        source_property (str): The property of the source entities to match on
        source_value_type (str, optional): The value type of the source property. Defaults to "single_value_str".
        target_class (str | None, optional): The class of the target entities. Defaults to None.
        target_property (str | None, optional): The property of the target entities to match on. Defaults to None.
        relationship_name (str, optional): The name of the relationship to add between the matched entities.
        link_direction (str, optional): The direction of the relationship. Defaults to "target_to_source".
        matching_method (str, optional): The matching method to use. Defaults to "regexp".
    Returns:
        int: The number of new links added to the graph
    """
    if source_value_type == "multi_value_str":
        # Split the values and add them as separate triples
        query = f"""
                    SELECT DISTINCT ?source ?source_value
                    WHERE {{
                        ?source rdf:type neat:{source_class} .
                        ?source neat:{source_property} ?source_value .
                    }}
                """
        triples_to_remove = []
        r1 = cast(tuple, graph_store.query(query))
        result = list(r1)
        for row in result:
            val_split = row[1].split(",")
            if len(val_split) > 1:
                triples_to_remove.append((row[0], URIRef(link_namespace + source_property), row[1]))
                for val in val_split:
                    graph_store.graph.add((row[0], URIRef(link_namespace + source_property), Literal(val)))

        for triple in triples_to_remove:
            graph_store.graph.remove(triple)

    graph_store.graph.commit()
    query = ""
    if matching_method == "exact_match":
        query = f"""
                    SELECT DISTINCT ?source ?target
                    WHERE {{
                        ?source rdf:type neat:{source_class} .
                        ?source neat:{source_property} ?source_value .
                        ?target rdf:type neat:{target_class} .
                        ?target neat:{target_property} ?target_value .
                        FILTER (?source_value = ?target_value)
                    }}
                """
    elif matching_method == "regexp":
        query = f"""
                    SELECT DISTINCT ?source ?target
                    WHERE {{
                        ?source rdf:type neat:{source_class} .
                        ?source neat:{source_property} ?source_value .
                        ?target rdf:type neat:{target_class} .
                        ?target neat:{target_property} ?target_value .
                        FILTER regex(?target_value,?source_value, "i")
                    }}
                """
    else:
        logging.error(f"Unknown matching method {matching_method}")
        return 0
    logging.debug(f"Running matcher query {query}")
    r1 = cast(tuple, graph_store.query(query))
    result = list(r1)
    logging.debug(f"Identified {len(result)} matches from the graph")
    new_links_counter = 0
    for row in result:
        new_links_counter += 1
        if link_direction == "target_to_source":
            graph_store.graph.add((row[1], URIRef(link_namespace + relationship_name), row[0]))
        else:
            graph_store.graph.add((row[0], URIRef(link_namespace + relationship_name), row[1]))

    return new_links_counter



import math
import re
import sys
import warnings
from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

from pydantic import Field, field_serializer, field_validator, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo
from rdflib import Namespace

import cognite.neat.rules.issues.spreadsheet
from cognite.neat.constants import PREFIXES
from cognite.neat.rules import exceptions
from cognite.neat.rules.models.rdfpath import (
    AllReferences,
    Hop,
    RawLookup,
    SingleProperty,
    SPARQLQuery,
    TransformationRuleType,
    Traversal,
    parse_rule,
)

from ._types import (
    ClassEntity,
    ContainerEntity,
    Entity,
    EntityTypes,
    NamespaceType,
    ParentClassEntity,
    ParentClassType,
    PrefixType,
    PropertyType,
    ReferenceEntity,
    ReferenceType,
    SemanticValueType,
    StrListType,
    Undefined,
    VersionType,
    ViewEntity,
    ViewPropEntity,
    XSDValueType,
)
from ._types._base import Unknown
from .base import (
    BaseMetadata,
    ExtensionCategory,
    ExtensionCategoryType,
    MatchType,
    RoleTypes,
    RuleModel,
    SchemaCompleteness,
    SheetEntity,
    SheetList,
)
from .domain_rules import DomainRules

if TYPE_CHECKING:
    from .dms_architect_rules import DMSProperty, DMSRules


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class InformationMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.information_architect
    schema_: SchemaCompleteness = Field(alias="schema")
    extension: ExtensionCategoryType | None = ExtensionCategory.addition
    prefix: PrefixType
    namespace: NamespaceType

    name: str = Field(
        alias="title",
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, min_length=1, max_length=1024)
    version: VersionType

    created: datetime = Field(
        description=("Date of the data model creation"),
    )

    updated: datetime = Field(
        description=("Date of the data model update"),
    )
    creator: StrListType = Field(
        description=(
            "List of contributors to the data model creation, "
            "typically information architects are considered as contributors."
        ),
    )

    @model_validator(mode="after")
    def extension_none_but_schema_extend(self) -> Self:
        if self.extension is None:
            self.extension = ExtensionCategory.addition
            return self
        return self


class InformationClass(SheetEntity):
    """
    Class is a category of things that share a common set of attributes and relationships.

    Args:
        class_: The class ID of the class.
        description: A description of the class.
        parent: The parent class of the class.
        reference: Reference of the source of the information for given resource
        match_type: The match type of the resource being described and the source entity.
    """

    parent: ParentClassType = Field(alias="Parent Class", default=None)
    reference: ReferenceType = Field(alias="Reference", default=None)
    match_type: MatchType | None = Field(alias="Match Type", default=None)
    comment: str | None = Field(alias="Comment", default=None)


class InformationProperty(SheetEntity):
    """
    A property is a characteristic of a class. It is a named attribute of a class that describes a range of values
    or a relationship to another class.

    Args:
        class_: Class ID to which property belongs
        property_: Property ID of the property
        name: Property name.
        value_type: Type of value property will hold (data or link to another class)
        min_count: Minimum count of the property values. Defaults to 0
        max_count: Maximum count of the property values. Defaults to None
        default: Default value of the property
        reference: Reference to the source of the information, HTTP URI
        match_type: The match type of the resource being described and the source entity.
        rule_type: Rule type for the transformation from source to target representation
                   of knowledge graph. Defaults to None (no transformation)
        rule: Actual rule for the transformation from source to target representation of
              knowledge graph. Defaults to None (no transformation)
    """

    property_: PropertyType = Field(alias="Property")
    value_type: SemanticValueType = Field(alias="Value Type")
    min_count: int | None = Field(alias="Min Count", default=None)
    max_count: int | float | None = Field(alias="Max Count", default=None)
    default: Any | None = Field(alias="Default", default=None)
    reference: ReferenceType = Field(alias="Reference", default=None)
    match_type: MatchType | None = Field(alias="Match Type", default=None)
    rule_type: str | TransformationRuleType | None = Field(alias="Rule Type", default=None)
    rule: str | AllReferences | SingleProperty | Hop | RawLookup | SPARQLQuery | Traversal | None = Field(
        alias="Rule", default=None
    )
    comment: str | None = Field(alias="Comment", default=None)

    @field_serializer("max_count", when_used="json-unless-none")
    def serialize_max_count(self, value: int | float | None) -> int | float | None | str:
        if isinstance(value, float) and math.isinf(value):
            return None
        return value

    @field_validator("max_count", mode="before")
    def parse_max_count(cls, value: int | float | None) -> int | float | None:
        if value is None:
            return float("inf")
        return value

    @model_validator(mode="after")
    def is_valid_rule(self):
        # TODO: Can we skip rule_type and simply try to parse the rule and if it fails, raise an error?
        if self.rule_type:
            self.rule_type = self.rule_type.lower()
            if not self.rule:
                raise exceptions.RuleTypeProvidedButRuleMissing(
                    self.property_, self.class_, self.rule_type
                ).to_pydantic_custom_error()
            self.rule = parse_rule(self.rule, self.rule_type)
        return self

    @model_validator(mode="after")
    def set_default_as_list(self):
        if (
            self.type_ == EntityTypes.data_property
            and self.default
            and self.is_list
            and not isinstance(self.default, list)
        ):
            if isinstance(self.default, str):
                if self.default:
                    self.default = self.default.replace(", ", ",").split(",")
                else:
                    self.default = [self.default]
        return self

    @model_validator(mode="after")
    def set_type_for_default(self):
        if self.type_ == EntityTypes.data_property and self.default:
            default_value = self.default[0] if isinstance(self.default, list) else self.default

            if type(default_value) != self.value_type.python:
                try:
                    if isinstance(self.default, list):
                        updated_list = []
                        for value in self.default:
                            updated_list.append(self.value_type.python(value))
                        self.default = updated_list
                    else:
                        self.default = self.value_type.python(self.default)

                except Exception:
                    exceptions.DefaultValueTypeNotProper(
                        self.property_,
                        type(self.default),
                        self.value_type.python,
                    )
        return self

    @property
    def type_(self) -> EntityTypes:
        """Type of property based on value type. Either data (attribute) or object (edge) property."""
        if self.value_type.type_ == EntityTypes.xsd_value_type:
            return EntityTypes.data_property
        elif self.value_type.type_ == EntityTypes.class_:
            return EntityTypes.object_property
        else:
            return EntityTypes.undefined

    @property
    def is_mandatory(self) -> bool:
        """Returns True if property is mandatory."""
        return self.min_count != 0

    @property
    def is_list(self) -> bool:
        """Returns True if property contains a list of values."""
        return self.max_count != 1


class InformationRules(RuleModel):
    metadata: InformationMetadata = Field(alias="Metadata")
    properties: SheetList[InformationProperty] = Field(alias="Properties")
    classes: SheetList[InformationClass] = Field(alias="Classes")
    prefixes: dict[str, Namespace] = Field(default_factory=lambda: PREFIXES.copy())
    reference: "InformationRules | None" = Field(None, alias="Reference")
    is_reference: bool = False

    @field_validator("prefixes", mode="before")
    def parse_str(cls, values: Any) -> Any:
        if isinstance(values, dict):
            return {key: Namespace(value) if isinstance(value, str) else value for key, value in values.items()}
        return values

    @model_validator(mode="after")
    def update_entities_prefix(self) -> Self:
        # update expected_value_types
        for property_ in self.properties:
            if property_.value_type.prefix is Undefined and property_.value_type.suffix is not Unknown:
                property_.value_type.prefix = self.metadata.prefix
            if property_.class_.prefix is Undefined:
                property_.class_.prefix = self.metadata.prefix

        # update parent classes
        for class_ in self.classes:
            if class_.parent:
                for parent in cast(list[ParentClassEntity], class_.parent):
                    if parent.prefix is Undefined:
                        parent.prefix = self.metadata.prefix
            if class_.class_.prefix is Undefined:
                class_.class_.prefix = self.metadata.prefix

        return self

    @model_validator(mode="after")
    def validate_schema_completeness(self) -> Self:
        # update expected_value_types

        if self.metadata.schema_ == SchemaCompleteness.complete:
            defined_classes = {class_.class_.versioned_id for class_ in self.classes}
            referred_classes = {property_.class_.versioned_id for property_ in self.properties} | {
                parent.versioned_id for class_ in self.classes for parent in class_.parent or []
            }
            referred_types = {
                property_.value_type.versioned_id
                for property_ in self.properties
                if property_.type_ == EntityTypes.object_property
                and not (isinstance(property_.value_type, Entity) and property_.value_type.suffix is Unknown)
            }
            if not referred_classes.issubset(defined_classes) or not referred_types.issubset(defined_classes):
                missing_classes = referred_classes.difference(defined_classes).union(
                    referred_types.difference(defined_classes)
                )
                raise exceptions.IncompleteSchema(missing_classes).to_pydantic_custom_error()

        return self

    @model_validator(mode="after")
    def validate_class_has_properties_or_parent(self) -> Self:
        defined_classes = {class_.class_ for class_ in self.classes if class_.reference is None}
        referred_classes = {property_.class_ for property_ in self.properties}
        has_parent_classes = {class_.class_ for class_ in self.classes if class_.parent}
        missing_classes = defined_classes.difference(referred_classes) - has_parent_classes
        if missing_classes:
            warnings.warn(
                cognite.neat.rules.issues.spreadsheet.ClassNoPropertiesNoParentsWarning(
                    [missing.versioned_id for missing in missing_classes]
                ),
                stacklevel=2,
            )
        return self

    @model_serializer(mode="plain", when_used="always")
    def information_rules_serializer(self, info: SerializationInfo) -> dict[str, Any]:
        kwargs = vars(info)
        default_prefix = f"{self.metadata.prefix}:" if self.metadata.prefix else ""

        field_names = ["Class", "Value Type"] if info.by_alias else ["class_", "value_type"]
        properties = []
        for prop in self.properties:
            dumped = prop.model_dump(**kwargs)
            for field_name in field_names:
                if value := dumped.get(field_name):
                    dumped[field_name] = value.removeprefix(default_prefix)
            properties.append(dumped)

        field_names = ["Class"] if info.by_alias else ["class_"]
        classes = []
        parent_name = "Parent Class" if info.by_alias else "parent"
        for cls in self.classes:
            dumped = cls.model_dump(**kwargs)
            for field_name in field_names:
                if value := dumped.get(field_name):
                    dumped[field_name] = value.removeprefix(default_prefix)
            if value := dumped.get(parent_name):
                dumped[parent_name] = ",".join(
                    constraint.strip().removeprefix(default_prefix) for constraint in value.split(",")
                )
            classes.append(dumped)

        return {
            "Metadata" if info.by_alias else "metadata": self.metadata.model_dump(**kwargs),
            "Classes" if info.by_alias else "classes": classes,
            "Properties" if info.by_alias else "properties": properties,
            "prefixes": {key: str(value) for key, value in self.prefixes.items()},
        }

    def as_domain_rules(self) -> DomainRules:
        return _InformationRulesConverter(self).as_domain_rules()

    def as_dms_architect_rules(self) -> "DMSRules":
        return _InformationRulesConverter(self).as_dms_architect_rules()

    def reference_self(self) -> "InformationRules":
        new_self = self.copy(deep=True)
        for prop in new_self.properties:
            prop.reference = ReferenceEntity(
                prefix=prop.class_.prefix,
                suffix=prop.class_.suffix,
                version=prop.class_.version,
                property_=prop.property_,
            )

        for cls_ in new_self.classes:
            cls_.reference = ReferenceEntity(
                prefix=cls_.class_.prefix, suffix=cls_.class_.suffix, version=cls_.class_.version
            )

        return new_self


class _InformationRulesConverter:
    def __init__(self, information: InformationRules):
        self.information = information

    def as_domain_rules(self) -> DomainRules:
        raise NotImplementedError("DomainRules not implemented yet")

    def as_dms_architect_rules(self, created: datetime | None = None, updated: datetime | None = None) -> "DMSRules":
        from .dms_architect_rules import DMSContainer, DMSMetadata, DMSProperty, DMSRules, DMSView

        info_metadata = self.information.metadata

        space = self._to_space(info_metadata.prefix)

        metadata = DMSMetadata(
            schema_=info_metadata.schema_,
            space=space,
            version=info_metadata.version,
            external_id=info_metadata.name.replace(" ", "_").lower(),
            creator=info_metadata.creator,
            name=info_metadata.name,
            created=created or datetime.now(),
            updated=updated or datetime.now(),
        )

        properties_by_class: dict[str, list[DMSProperty]] = defaultdict(list)
        for prop in self.information.properties:
            properties_by_class[prop.class_.versioned_id].append(self._as_dms_property(prop))

        views: list[DMSView] = [
            DMSView(
                class_=cls_.class_,
                view=ViewPropEntity(prefix=cls_.class_.prefix, suffix=cls_.class_.suffix, version=cls_.class_.version),
                description=cls_.description,
                reference=cls_.reference,
                implements=self._get_view_implements(cls_, info_metadata),
            )
            for cls_ in self.information.classes
        ]

        classes_without_properties: set[str] = set()
        for class_ in self.information.classes:
            properties: list[DMSProperty] = properties_by_class.get(class_.class_.versioned_id, [])
            if not properties or all(
                isinstance(prop.value_type, ViewPropEntity) and prop.relation != "direct" for prop in properties
            ):
                classes_without_properties.add(class_.class_.versioned_id)

        containers: list[DMSContainer] = []
        for class_ in self.information.classes:
            if class_.class_.versioned_id in classes_without_properties:
                continue
            containers.append(
                DMSContainer(
                    class_=class_.class_,
                    container=ContainerEntity(prefix=class_.class_.prefix, suffix=class_.class_.suffix),
                    description=class_.description,
                    constraint=[
                        ContainerEntity(
                            prefix=parent.prefix,
                            suffix=parent.suffix,
                        )
                        for parent in class_.parent or []
                        if parent.versioned_id not in classes_without_properties
                    ]
                    or None,
                )
            )

        return DMSRules(
            metadata=metadata,
            properties=SheetList[DMSProperty](
                data=[prop for prop_set in properties_by_class.values() for prop in prop_set]
            ),
            views=SheetList[DMSView](data=views),
            containers=SheetList[DMSContainer](data=containers),
            reference=self.information.reference and self.information.reference.as_dms_architect_rules(),  # type: ignore[arg-type]
            is_reference=self.information.is_reference,
        )

    @classmethod
    def _as_dms_property(cls, prop: InformationProperty) -> "DMSProperty":
        """This creates the first"""

        from .dms_architect_rules import DMSProperty

        # returns property type, which can be ObjectProperty or DatatypeProperty
        if isinstance(prop.value_type, XSDValueType):
            value_type = cast(XSDValueType, prop.value_type).dms._type.casefold()  # type: ignore[attr-defined]
        elif isinstance(prop.value_type, ClassEntity):
            value_type = ViewPropEntity(
                prefix=prop.value_type.prefix, suffix=prop.value_type.suffix, version=prop.value_type.version
            )
        else:
            raise ValueError(f"Unsupported value type: {prop.value_type.type_}")

        relation: Literal["direct", "multiedge"] | None = None
        if isinstance(value_type, ViewPropEntity):
            relation = "multiedge" if prop.is_list else "direct"

        container: ContainerEntity | None = None
        container_property: str | None = None
        is_list: bool | None = prop.is_list
        nullable: bool | None = not prop.is_mandatory
        if relation == "multiedge":
            is_list = None
            nullable = None
        elif relation == "direct":
            nullable = True
            container, container_property = cls._get_container(prop)
        else:
            container, container_property = cls._get_container(prop)

        return DMSProperty(
            class_=prop.class_,
            property_=prop.property_,
            value_type=value_type,
            nullable=nullable,
            is_list=is_list,
            relation=relation,
            default=prop.default,
            reference=prop.reference,
            container=container,
            container_property=container_property,
            view=ViewPropEntity(prefix=prop.class_.prefix, suffix=prop.class_.suffix, version=prop.class_.version),
            view_property=prop.property_,
        )

    @classmethod
    def _to_space(cls, prefix: str) -> str:
        """Ensures that the prefix comply with the CDF space regex"""
        prefix = re.sub(r"[^a-zA-Z0-9_-]", "_", prefix)
        if prefix[0].isdigit() or prefix[0] == "_":
            prefix = f"a{prefix}"
        prefix = prefix[:43]
        if prefix[-1] == "_":
            prefix = f"{prefix[:-1]}1"
        return prefix

    @classmethod
    def _get_container(cls, prop: InformationProperty) -> tuple[ContainerEntity, str]:
        if isinstance(prop.reference, ReferenceEntity):
            return (
                ContainerEntity(prefix=prop.reference.prefix, suffix=prop.reference.suffix),
                prop.reference.property_ or prop.property_,
            )
        else:
            return ContainerEntity(prefix=prop.class_.prefix, suffix=prop.class_.suffix), prop.property_

    @classmethod
    def _get_view_implements(cls, cls_: InformationClass, metadata: InformationMetadata) -> list[ViewEntity]:
        if isinstance(cls_.reference, ReferenceEntity) and cls_.reference.prefix != metadata.prefix:
            # We use the reference for implements if it is in a different namespace
            implements = [
                ViewEntity(prefix=cls_.reference.prefix, suffix=cls_.reference.suffix, version=cls_.reference.version)
            ]
        else:
            implements = []

        implements.extend(
            [
                ViewEntity(prefix=parent.prefix, suffix=parent.suffix, version=parent.version)
                for parent in cls_.parent or []
            ]
        )
        return implements
